# lingo - payment and billing system
# Copyright (C) 2022  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import copy
import datetime

from django.db import transaction
from django.test.client import RequestFactory
from django.utils import formats
from django.utils.timezone import localtime
from django.utils.translation import gettext_lazy as _

from lingo.agendas.chrono import get_check_status, get_subscriptions
from lingo.agendas.models import Agenda, CheckType
from lingo.invoicing.models import (
    DraftInvoice,
    DraftInvoiceLine,
    DraftJournalLine,
    InjectedLine,
    InvoicingError,
    Payer,
    Regie,
)
from lingo.pricing.models import Pricing, PricingError, PricingNotFound


def get_agendas(pool):
    agendas_pricings = Pricing.objects.filter(
        flat_fee_schedule=False, agendas__regie=pool.campaign.regie, agendas__in=pool.campaign.agendas.all()
    ).extra(
        where=['(date_start, date_end) OVERLAPS (%s, %s)'],
        params=[pool.campaign.date_start, pool.campaign.date_end],
    )
    return Agenda.objects.filter(pk__in=agendas_pricings.values('agendas')).order_by('pk')


def get_users_from_subscriptions(agendas, pool):
    users = {}
    for agenda in agendas:
        subscriptions = get_subscriptions(
            agenda_slug=agenda.slug,
            date_start=pool.campaign.date_start,
            date_end=pool.campaign.date_end,
        )
        for subscription in subscriptions:
            user_external_id = subscription['user_external_id']
            if user_external_id in users:
                continue
            users[user_external_id] = (subscription['user_first_name'], subscription['user_last_name'])
    return list(users.items())


def get_pricing(agendas_pricings_for_agenda, date_event):
    # same logic as Pricing.get_pricing
    for pricing in agendas_pricings_for_agenda:
        if pricing.date_start > date_event:
            continue
        if pricing.date_end <= date_event:
            continue
        return pricing
    raise PricingNotFound


def get_cached_payer_data(request, pool, payer_data_cache, payer_external_id, pricing=None, payer_data=None):
    if payer_external_id not in payer_data_cache:
        if pricing:
            # will raise a PricingError if payer_data can not be computed
            payer_data_cache[payer_external_id] = pool.campaign.regie.get_payer_data(
                request, payer_external_id
            )
        elif payer_data:
            payer_data_cache[payer_external_id] = payer_data
    return payer_data_cache.get(payer_external_id) or {}


def get_invoice_lines_for_user(
    agendas,
    agendas_pricings,
    user_external_id,
    user_first_name,
    user_last_name,
    pool,
    payer_data_cache,
):
    if not agendas:
        return []

    agendas_by_slug = {a.slug: a for a in agendas}
    agendas_pricings_by_agendas = collections.defaultdict(list)
    for pricing in agendas_pricings:
        if pricing.flat_fee_schedule:
            continue
        for agenda in pricing.agendas.all():
            agendas_pricings_by_agendas[agenda.slug].append(pricing)
    check_types = {(c.slug, c.group.slug, c.kind): c for c in CheckType.objects.select_related('group').all()}

    # get check status for user_external_id, on agendas, for the period
    check_status_list = get_check_status(
        agenda_slugs=list(agendas_by_slug.keys()),
        user_external_id=user_external_id,
        date_start=pool.campaign.date_start,
        date_end=pool.campaign.date_end,
    )
    request = RequestFactory().get('/')  # XXX

    # get resolved errors in last pool for this user
    previous_pool = pool.campaign.pool_set.exclude(pk=pool.pk).order_by('created_at').last()
    resolved_errors = (
        {
            li.slug: li
            for li in DraftJournalLine.objects.filter(
                pool=previous_pool, status='error', user_external_id=user_external_id
            ).exclude(error_status='')
        }
        if previous_pool is not None
        else {}
    )

    # build lines from check status
    lines = []
    for check_status in check_status_list:
        serialized_event = check_status['event']
        serialized_booking = check_status['booking']
        event_date = datetime.datetime.fromisoformat(serialized_event['start_datetime']).date()
        event_slug = '%s@%s' % (serialized_event['agenda'], serialized_event['slug'])

        agenda = agendas_by_slug[serialized_event['agenda']]
        quantity = 1
        quantity_type = 'units'
        if agenda.partial_bookings:
            quantity = serialized_booking.get('computed_duration') or 0
            quantity_type = 'minutes'
        payer_external_id = _('unknown')
        payer_data = {}
        try:
            pricing = get_pricing(agendas_pricings_by_agendas.get(agenda.slug), event_date)
            payer_external_id = pool.campaign.regie.get_payer_external_id(request, user_external_id)
            payer_data = get_cached_payer_data(
                request, pool, payer_data_cache, payer_external_id, pricing=pricing
            )
            pricing_data = pricing.get_pricing_data_for_event(
                request=request,
                agenda=agenda,
                event=serialized_event,
                check_status=check_status['check_status'],
                user_external_id=user_external_id,
                payer_external_id=payer_external_id,
            )

            line_kwargs = {
                'event_date': event_date,
                'slug': event_slug,
                'quantity_type': quantity_type,
                'user_external_id': user_external_id,
                'user_first_name': user_first_name,
                'user_last_name': user_last_name,
                'payer_external_id': payer_external_id,
                'payer_first_name': payer_data['first_name'],
                'payer_last_name': payer_data['last_name'],
                'payer_address': payer_data['address'],
                'payer_demat': payer_data['demat'],
                'payer_direct_debit': payer_data['direct_debit'],
                'booking': serialized_booking,
                'status': 'success',
                'pool': pool,
            }
            booking_details = pricing_data.get('booking_details') or {}
            if agenda.partial_bookings and booking_details.get('status') in ['absence', 'presence']:
                # partial booking: we need to first invoice booked hours, then apply overtaking or reductions
                check_type = check_types.get(
                    (
                        booking_details.get('check_type'),
                        booking_details.get('check_type_group'),
                        booking_details.get('status'),
                    )
                )
                normal_pricing_data = pricing_data
                if booking_details['status'] != 'presence' or booking_details.get('check_type'):
                    # calculate pricing for absence without check type
                    normal_pricing_data = pricing.get_pricing_data_for_event(
                        request=request,
                        agenda=agenda,
                        event=serialized_event,
                        check_status={'status': 'presence', 'check_type': None},
                        user_external_id=user_external_id,
                        payer_external_id=payer_external_id,
                    )
                normal_pricing = normal_pricing_data['pricing']
                current_pricing = pricing_data['pricing']
                if serialized_booking.get('adjusted_duration'):
                    # there was a booking,
                    # first add line for booked hours
                    lines.append(
                        DraftJournalLine(
                            label=agenda.label,
                            description='@booked-hours@',
                            amount=normal_pricing,
                            quantity=serialized_booking['adjusted_duration'],
                            pricing_data=normal_pricing_data,
                            event=serialized_event,
                            **line_kwargs,
                        )
                    )
                    # and add overtaking if needed
                    if serialized_booking['computed_duration'] > serialized_booking['adjusted_duration']:
                        overtaking_event = copy.deepcopy(serialized_event)
                        if overtaking_event.get('primary_event'):
                            overtaking_event['primary_event'] += '::overtaking'
                        lines.append(
                            DraftJournalLine(
                                label=_('Overtaking'),
                                description='@overtaking@',  # so days will be not displayed
                                amount=normal_pricing,
                                quantity=serialized_booking['computed_duration']
                                - serialized_booking['adjusted_duration'],
                                pricing_data=normal_pricing_data,
                                event=overtaking_event,
                                **line_kwargs,
                            )
                        )
                    diff_pricing = current_pricing - normal_pricing
                    if diff_pricing != 0:
                        # reduction of overcharging to apply
                        label = ''
                        if check_type:
                            label = check_type.label
                        elif booking_details['status'] == 'absence':
                            label = _('Absence')
                        diff_event = copy.deepcopy(serialized_event)
                        if diff_event.get('primary_event'):
                            diff_event['primary_event'] += ':%s:%s' % (
                                booking_details['status'],
                                booking_details.get('check_type') or '',
                            )
                        lines.append(
                            DraftJournalLine(
                                label=label,
                                # if negative pricing, quantity is negative and pricing is set to the opposite
                                amount=abs(diff_pricing),
                                quantity=-serialized_booking['computed_duration']
                                if diff_pricing < 0
                                else serialized_booking['computed_duration'],
                                pricing_data=pricing_data,
                                event=diff_event,
                                **line_kwargs,
                            )
                        )
                else:
                    # no booking
                    label = serialized_event['label']
                    if check_type:
                        label = check_type.label
                    elif booking_details.get('status') == 'presence':
                        label = _('Presence without booking')
                    lines.append(
                        DraftJournalLine(
                            label=label,
                            amount=pricing_data['pricing'],
                            quantity=quantity,
                            pricing_data=pricing_data,
                            event=serialized_event,
                            **line_kwargs,
                        )
                    )

            else:
                # not partial bookings
                lines.append(
                    DraftJournalLine(
                        label=serialized_event['label'],
                        amount=pricing_data['pricing'],
                        quantity=quantity,
                        pricing_data=pricing_data,
                        event=serialized_event,
                        **line_kwargs,
                    )
                )
        except (PricingError, InvoicingError) as e:
            # PricingNotFound: can happen if pricing model defined only on a part of the requested period
            pricing_error = {
                'error': type(e).__name__,
                'error_details': e.details,
            }
            error_status = ''
            if resolved_errors.get(event_slug):
                error_status = resolved_errors[event_slug].error_status
            lines.append(
                DraftJournalLine(
                    event_date=event_date,
                    slug=event_slug,
                    label=serialized_event['label'],
                    amount=0,
                    quantity=quantity,
                    quantity_type=quantity_type,
                    user_external_id=user_external_id,
                    user_first_name=user_first_name,
                    user_last_name=user_last_name,
                    payer_external_id=payer_external_id,
                    payer_first_name=payer_data.get('first_name') or '',
                    payer_last_name=payer_data.get('last_name') or '',
                    payer_address=payer_data.get('address') or '',
                    payer_demat=payer_data.get('demat') or False,
                    payer_direct_debit=payer_data.get('direct_debit') or False,
                    event=serialized_event,
                    booking=serialized_booking,
                    pricing_data=pricing_error,
                    status='warning' if isinstance(e, PricingNotFound) else 'error',
                    error_status=error_status,
                    pool=pool,
                )
            )

    if pool.campaign.injected_lines != 'no':
        # fetch injected lines
        injected_lines = (
            InjectedLine.objects.filter(
                regie=pool.campaign.regie,
                event_date__lt=pool.campaign.date_end,
                user_external_id=user_external_id,
            )
            .exclude(
                # exclude already invoiced lines
                journalline__isnull=False
            )
            .exclude(
                # exclude lines used in another campaign
                pk__in=DraftJournalLine.objects.filter(from_injected_line__isnull=False)
                .exclude(pool__campaign=pool.campaign)
                .values('from_injected_line')
            )
            .order_by('event_date')
        )
        if pool.campaign.injected_lines == 'period':
            injected_lines = injected_lines.filter(
                event_date__gte=pool.campaign.date_start,
            )

        for injected_line in injected_lines:
            payer_external_id = injected_line.payer_external_id
            payer_data = {
                'first_name': injected_line.payer_first_name,
                'last_name': injected_line.payer_last_name,
                'address': injected_line.payer_address,
                'demat': injected_line.payer_demat,
                'direct_debit': injected_line.payer_direct_debit,
            }
            payer_data = get_cached_payer_data(
                request, pool, payer_data_cache, payer_external_id, payer_data=payer_data
            )
            lines.append(
                DraftJournalLine(
                    event_date=injected_line.event_date,
                    slug=injected_line.slug,
                    label=injected_line.label,
                    amount=injected_line.amount,
                    user_external_id=user_external_id,
                    user_first_name=user_first_name,
                    user_last_name=user_last_name,
                    payer_external_id=payer_external_id,
                    payer_first_name=payer_data['first_name'],
                    payer_last_name=payer_data['last_name'],
                    payer_address=payer_data['address'],
                    payer_demat=payer_data['demat'],
                    payer_direct_debit=payer_data['direct_debit'],
                    status='success',
                    pool=pool,
                    from_injected_line=injected_line,
                )
            )

    DraftJournalLine.objects.bulk_create(lines)

    return lines


def get_all_invoice_lines(agendas, users, pool):
    agendas_pricings = (
        Pricing.objects.filter(flat_fee_schedule=False)
        .extra(
            where=['(date_start, date_end) OVERLAPS (%s, %s)'],
            params=[pool.campaign.date_start, pool.campaign.date_end],
        )
        .prefetch_related('agendas', 'criterias', 'categories')
    )

    lines = []
    payer_data_cache = {}
    for user_external_id, (user_first_name, user_last_name) in users:
        # generate lines for each user
        lines += get_invoice_lines_for_user(
            agendas=agendas,
            agendas_pricings=agendas_pricings,
            user_external_id=user_external_id,
            user_first_name=user_first_name,
            user_last_name=user_last_name,
            pool=pool,
            payer_data_cache=payer_data_cache,
        )
    return lines


def generate_invoices_from_lines(all_lines, pool):
    regie = pool.campaign.regie

    # regroup lines by payer_external_id (payer)
    lines = {}
    for line in all_lines:
        if line.status != 'success':
            # ignore lines in error
            continue
        if line.payer_external_id not in lines:
            lines[line.payer_external_id] = {
                'payer_first_name': line.payer_first_name,
                'payer_last_name': line.payer_last_name,
                'payer_address': line.payer_address,
                'payer_demat': line.payer_demat,
                'payer_direct_debit': line.payer_direct_debit,
                'lines': [],
            }
        lines[line.payer_external_id]['lines'].append(line)

    def is_line_to_be_ignored(line):
        if line.pricing_data.get('booking_details', {}).get('status') == 'not-booked':
            # ignore not booked events
            return True
        return False

    # generate invoices by regie and by payer_external_id (payer)
    invoices = []
    agendas_by_slug = {a.slug: a for a in Agenda.objects.all()}
    check_types = {(c.slug, c.group.slug, c.kind): c for c in CheckType.objects.select_related('group').all()}
    for payer_external_id, payer_data in lines.items():
        has_line = False
        for line in payer_data['lines']:
            if is_line_to_be_ignored(line):
                continue
            has_line = True
            break
        if not has_line:
            # don't create an empty invoice
            continue
        invoice = DraftInvoice.objects.create(
            label=_('Invoice from %(start)s to %(end)s')
            % {
                'start': pool.campaign.date_start.strftime('%d/%m/%Y'),
                'end': (pool.campaign.date_end - datetime.timedelta(days=1)).strftime('%d/%m/%Y'),
            },
            date_publication=pool.campaign.date_publication,
            date_payment_deadline=pool.campaign.date_payment_deadline,
            date_due=pool.campaign.date_due,
            date_debit=pool.campaign.date_debit if payer_data['payer_direct_debit'] else None,
            regie=regie,
            payer_external_id=payer_external_id,
            payer_first_name=payer_data['payer_first_name'],
            payer_last_name=payer_data['payer_last_name'],
            payer_address=payer_data['payer_address'],
            payer_demat=payer_data['payer_demat'],
            payer_direct_debit=payer_data['payer_direct_debit'],
            pool=pool,
        )
        # regroup journal lines by user_external_id, status, check_type, check_type_group, pricing
        grouped_lines = collections.defaultdict(list)
        other_lines = []
        for line in payer_data['lines']:
            if is_line_to_be_ignored(line):
                continue
            if not line.event.get('primary_event'):
                # not a recurring event
                other_lines.append(line)
                continue
            key = (
                line.user_external_id,
                line.event['agenda'],
                line.event['primary_event'],
                line.pricing_data.get('booking_details', {}).get('status'),
                line.pricing_data.get('booking_details', {}).get('check_type'),
                line.pricing_data.get('booking_details', {}).get('check_type_group'),
                line.amount,
                line.quantity_type,
            )
            grouped_lines[key].append(line)
        for key, journal_lines in grouped_lines.items():
            journal_lines = sorted(journal_lines, key=lambda li: li.pk)
            first_line = journal_lines[0]
            slug = '-'.join(str(k) for k in key[1:])
            check_type = check_types.get((key[4], key[5], key[3]))
            event_datetime = localtime(
                datetime.datetime.fromisoformat(journal_lines[0].event['start_datetime'])
            )
            event_time = event_datetime.time().isoformat()
            quantity = sum(li.quantity for li in journal_lines)
            if key[7] == 'minutes':
                quantity = quantity / 60  # convert in hours
            agenda = agendas_by_slug.get(key[1])
            dates = sorted(li.event_date for li in journal_lines)
            description = ''
            if first_line.description:
                description = first_line.description
                if description == '@booked-hours@':
                    rounded_quantity = int(quantity)  # remove leading zero
                    if rounded_quantity != quantity:
                        rounded_quantity = quantity
                    description = _('%s booked hours for the period') % rounded_quantity
            elif dates:
                description = ', '.join(formats.date_format(d, 'Dd') for d in dates)
            invoice_line = DraftInvoiceLine.objects.create(
                invoice=invoice,
                event_date=pool.campaign.date_start,
                slug=slug,
                label=first_line.label,
                quantity=quantity,
                unit_amount=first_line.amount,
                details={
                    'agenda': key[1],
                    'primary_event': key[2],
                    'status': key[3],
                    'check_type': key[4],
                    'check_type_group': key[5],
                    'check_type_label': check_type.label if check_type else key[4],
                    'dates': dates,
                    'event_time': event_time,
                    'partial_bookings': agenda.partial_bookings if agenda else False,
                },
                event_slug='%s@%s' % (key[1], key[2]),
                event_label=first_line.event.get('label') or first_line.label,
                agenda_slug=key[1],
                activity_label=agenda.label if agenda else '',
                description=description,
                user_external_id=first_line.user_external_id,
                user_first_name=first_line.user_first_name,
                user_last_name=first_line.user_last_name,
                payer_external_id=payer_external_id,
                payer_first_name=payer_data['payer_first_name'],
                payer_last_name=payer_data['payer_last_name'],
                payer_address=payer_data['payer_address'],
                payer_demat=payer_data['payer_demat'],
                payer_direct_debit=payer_data['payer_direct_debit'],
                pool=pool,
            )
            DraftJournalLine.objects.filter(pk__in=[li.pk for li in journal_lines]).update(
                invoice_line=invoice_line
            )
        for line in other_lines:
            agenda_slug = ''
            if '@' in line.slug:
                agenda_slug = line.slug.split('@')[0]
            agenda = agendas_by_slug.get(agenda_slug)
            invoice_line = DraftInvoiceLine.objects.create(
                invoice=invoice,
                event_date=line.event_date,
                slug=line.slug,
                label=line.label,
                quantity=1,
                unit_amount=line.amount,
                event_slug=line.slug,
                event_label=line.event.get('label') or line.label,
                agenda_slug=agenda_slug,
                activity_label=agenda.label if agenda else '',
                user_external_id=line.user_external_id,
                user_first_name=line.user_first_name,
                user_last_name=line.user_last_name,
                payer_external_id=payer_external_id,
                payer_first_name=payer_data['payer_first_name'],
                payer_last_name=payer_data['payer_last_name'],
                payer_address=payer_data['payer_address'],
                payer_demat=payer_data['payer_demat'],
                payer_direct_debit=payer_data['payer_direct_debit'],
                pool=pool,
            )
            line.invoice_line = invoice_line
            line.save()
        invoices.append(invoice)

    return invoices


def export_site(
    regies=True,
    payers=True,
):
    '''Dump site objects to JSON-dumpable dictionnary'''
    data = {}
    if regies:
        data['regies'] = [x.export_json() for x in Regie.objects.all()]
    if payers:
        data['payers'] = [x.export_json() for x in Payer.objects.all()]
    return data


def import_site(data):
    results = {
        key: collections.defaultdict(list)
        for key in [
            'regies',
            'payers',
        ]
    }

    with transaction.atomic():
        for cls, key in ((Payer, 'payers'), (Regie, 'regies')):
            objs = data.get(key, [])
            for obj in objs:
                created, obj = cls.import_json(obj)
                results[key]['all'].append(obj)
                if created:
                    results[key]['created'].append(obj)
                else:
                    results[key]['updated'].append(obj)
    return results
