import eopayment
import pytest
from django.urls import reverse

from lingo.epayment.models import PaymentBackend, Transaction
from lingo.invoicing.models import Regie
from tests.utils import login

pytestmark = pytest.mark.django_db


def test_manager_epayment_transaction_list_empty(app, admin_user):
    app = login(app)
    resp = app.get(reverse('lingo-manager-epayment-transaction-list'))
    h2 = resp.pyquery('div#appbar h2')
    assert h2.text() == 'E-Payment'
    assert 'There are no transactions yet.'


def test_manager_epayment_transaction_list_status_filter(app, admin_user):
    Transaction.objects.create(
        order_id='1234', bank_transaction_id='2345', status=eopayment.WAITING, amount=20
    )
    Transaction.objects.create(order_id='abcd', bank_transaction_id='bcde', status=eopayment.PAID, amount=20)
    Transaction.objects.create(order_id='ABCD', bank_transaction_id='BCDE', status=eopayment.ERROR, amount=20)
    app = login(app)
    resp = app.get(reverse('lingo-manager-epayment-transaction-list'))
    assert resp.pyquery('tbody tr').length == 3

    resp.form['status'] = 'running'
    resp = resp.form.submit()
    assert resp.pyquery('tbody tr').length == 1
    assert '1234' in resp.pyquery('tbody tr').text()

    resp.form['status'] = 'paid'
    resp = resp.form.submit()
    assert resp.pyquery('tbody tr').length == 1
    assert 'abcd' in resp.pyquery('tbody tr').text()

    resp.form['status'] = 'others'
    resp = resp.form.submit()
    assert resp.pyquery('tbody tr').length == 1
    assert 'ABCD' in resp.pyquery('tbody tr').text()

    resp.form['status'] = ''
    resp = resp.form.submit()
    assert resp.pyquery('tbody tr').length == 3


def test_manager_epayment_backend_list_empty(app, admin_user):
    app = login(app)
    resp = app.get(reverse('lingo-manager-epayment-transaction-list'))
    resp = resp.click('Configure payment backends')
    h2 = resp.pyquery('div#appbar h2')
    assert h2.text() == 'Payment backends'
    assert 'There are no payment backend yet.' in resp.text


def test_manager_epayment_backend_add(app, admin_user):
    regie = Regie.objects.create(label='Bar')
    app = login(app)
    resp = app.get(reverse('lingo-manager-epayment-backend-list'))
    resp = resp.click('New payment backend')
    resp.form.set('label', 'Test')
    resp.form.set('service', 'dummy')
    resp = resp.form.submit().follow()
    assert 'Please fill additional backend parameters.' in resp.text
    resp.form.set('regie', regie.id)
    resp.form.set('origin', 'blah')
    resp = resp.form.submit().follow()
    assert resp.pyquery('.objects-list li').length == 1
    backend = PaymentBackend.objects.all().first()
    assert backend.slug == 'test'
    assert backend.service_options['origin'] == 'blah'


def test_manager_epayment_backend_edit(app, admin_user):
    regie = Regie.objects.create(label='Bar')
    backend = PaymentBackend.objects.create(
        label='Test', slug='test', service='dummy', service_options={'origin': 'Blah'}, regie=regie
    )
    app = login(app)
    resp = app.get(reverse('lingo-manager-epayment-backend-edit', kwargs={'pk': backend.pk}))
    resp.form.set('origin', 'change')
    resp = resp.form.submit().follow()
    backend.refresh_from_db()
    assert backend.service_options['origin'] == 'change'


def test_manager_epayment_backend_delete(app, admin_user):
    backend = PaymentBackend.objects.create(label='Test', slug='test', service='dummy')
    app = login(app)
    resp = app.get(reverse('lingo-manager-epayment-backend-delete', kwargs={'pk': backend.pk}))
    resp = resp.form.submit().follow()
    assert PaymentBackend.objects.all().count() == 0
