import datetime
import urllib
from unittest import mock

import pytest
from django.urls import reverse
from django.utils.timezone import now

from lingo.basket.models import Basket, BasketLine, BasketLineItem
from lingo.epayment.models import PaymentBackend, Transaction
from lingo.invoicing.models import (
    Credit,
    CreditAssignment,
    CreditLine,
    DraftInvoice,
    DraftInvoiceLine,
    Payment,
    PaymentType,
    Regie,
)
from tests.utils import login

pytestmark = pytest.mark.django_db


def test_payment_disabled_demo(app, settings):
    settings.DEBUG = False
    app.get(reverse('lingo-epayment-demo'), status=404)


def test_payment_redirect(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    resp = app.get(reverse('lingo-epayment-demo'))
    assert resp.location.startswith('https://dummy-payment.entrouvert.com/?')
    assert Transaction.objects.all().count() == 1
    transaction = Transaction.objects.all().first()
    assert transaction.status == 0
    assert transaction.amount == 20


def test_payment_redirect_select_backend(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    resp = app.get(reverse('lingo-epayment-demo') + '?backend=test')
    assert resp.location.startswith('https://dummy-payment.entrouvert.com/?')
    assert Transaction.objects.all().count() == 1
    transaction = Transaction.objects.all().first()
    assert transaction.status == 0
    assert transaction.amount == 20


def test_payment_redirect_by_form(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(
        label='Test',
        service='ogone',
        service_options={'environment': 'TEST', 'pspid': 'xxx', 'sha_in': 'xxx', 'sha_out': 'xxx'},
    )
    resp = app.get(reverse('lingo-epayment-demo'))
    assert resp.pyquery('input[name="AMOUNT"]').attr.value == '2000'
    assert Transaction.objects.all().count() == 1
    transaction = Transaction.objects.all().first()
    assert transaction.status == 0
    assert transaction.amount == 20


def test_payment_return(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    app.get(reverse('lingo-epayment-demo'))  # init transaction
    transaction = Transaction.objects.all().first()
    resp = app.get(
        reverse('lingo-epayment-explicit-return', kwargs={'transaction_id': transaction.id})
        + f'?transaction_id={transaction.order_id}&origin=origin&ok=1&signed=1'
    ).follow()
    status_url = reverse('lingo-epayment-processing-status', kwargs={'transaction_id': transaction.id})
    assert status_url in resp
    assert resp.pyquery('#continue a').attr.href == '/'
    assert app.get(status_url).json == {'status': 3, 'running': False, 'paid': True}


def test_payment_return_next_url(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    app.get(reverse('lingo-epayment-demo') + '?next_url=/next/')  # init transaction
    transaction = Transaction.objects.all().first()
    resp = app.get(
        reverse('lingo-epayment-explicit-return', kwargs={'transaction_id': transaction.id})
        + f'?transaction_id={transaction.order_id}&origin=origin&ok=1&signed=1'
    ).follow()
    status_url = reverse('lingo-epayment-processing-status', kwargs={'transaction_id': transaction.id})
    assert status_url in resp
    assert resp.pyquery('#continue a').attr.href == '/next/'
    assert app.get(status_url).json == {'status': 3, 'running': False, 'paid': True}


def test_payment_callback(app, settings):
    settings.DEBUG = True
    PaymentBackend.objects.create(label='Test', service='dummy')
    app.get(reverse('lingo-epayment-demo'))  # init transaction
    transaction = Transaction.objects.all().first()
    app.get(
        reverse('lingo-epayment-explicit-callback', kwargs={'transaction_id': transaction.id})
        + f'?transaction_id={transaction.order_id}&origin=origin&ok=1&signed=1',
        status=200,
    )
    transaction.refresh_from_db()
    assert transaction.status == 3


def test_payment_auto_return(app, settings):
    settings.DEBUG = True
    app.get(reverse('lingo-epayment-auto-return'), status=400)

    backend = PaymentBackend.objects.create(
        label='Test',
        service='ogone',
        service_options={
            'environment': 'TEST',
            'pspid': 'xxx',
            'sha_in': 'xxx',
            'sha_out': 'xxx',
            'encoding': 'utf-8',
        },
    )
    app.get(reverse('lingo-epayment-demo'))  # init transaction
    transaction = Transaction.objects.all().first()

    ogone_backend = backend.eopayment
    data = {
        'orderid': transaction.order_id,
        'status': '9',
        'payid': '3011229363',
        'ncerror': '0',
    }
    data['shasign'] = ogone_backend.backend.sha_sign_out(data, encoding='utf-8')
    resp = app.get(reverse('lingo-epayment-auto-return') + '?' + urllib.parse.urlencode(data))
    assert resp.location == reverse('lingo-epayment-processing', kwargs={'transaction_id': transaction.id})
    transaction.refresh_from_db()
    assert transaction.status == 3


def test_payment_auto_callback(app, settings):
    settings.DEBUG = True
    assert app.get(reverse('lingo-epayment-auto-callback')).json['err'] == 1

    backend = PaymentBackend.objects.create(
        label='Test',
        service='ogone',
        service_options={
            'environment': 'TEST',
            'pspid': 'xxx',
            'sha_in': 'xxx',
            'sha_out': 'xxx',
            'encoding': 'utf-8',
        },
    )
    app.get(reverse('lingo-epayment-demo'))  # init transaction
    transaction = Transaction.objects.all().first()

    ogone_backend = backend.eopayment
    data = {
        'orderid': transaction.order_id,
        'status': '9',
        'payid': '3011229363',
        'ncerror': '0',
    }
    data['shasign'] = ogone_backend.backend.sha_sign_out(data, encoding='utf-8')
    assert (
        app.get(reverse('lingo-epayment-auto-callback') + '?' + urllib.parse.urlencode(data)).json['err'] == 0
    )
    transaction.refresh_from_db()
    assert transaction.status == 3


def test_basket_payment(app, simple_user):
    app = login(app, username='user', password='user')

    regie = Regie.objects.create(label='Foo')
    PaymentType.create_defaults(regie)
    PaymentBackend.objects.create(label='Test', service='dummy', regie=regie)
    invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    DraftInvoiceLine.objects.create(
        slug='event-a-foo-bar',
        label='Event A',
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice,
        quantity=10,
        unit_amount=1,
        user_external_id='user:1',
    )

    basket = Basket.objects.create(
        regie=regie,
        draft_invoice=invoice,
        payer_nameid='ab' * 16,
        expiry_at=now() + datetime.timedelta(hours=1),
    )
    line = BasketLine.objects.create(
        basket=basket,
        closed=False,
        user_external_id='user:1',
        payment_callback_url='http://payment1.com',
    )
    BasketLineItem.objects.create(
        line=line,
        label='Event A',
        subject='Réservation',
        details='Lun 06/11, Mar 07/11',
        quantity=10,
        unit_amount=1,
    )
    line.closed = True
    line.save()

    resp = app.get('/basket/validate/')
    resp = resp.form.submit()
    assert resp.location.startswith('https://dummy-payment.entrouvert.com/')
    assert Transaction.objects.all().count() == 1
    transaction = Transaction.objects.all().first()
    assert transaction.status == 0
    assert transaction.amount == 10

    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        resp = app.get(
            reverse('lingo-epayment-explicit-return', kwargs={'transaction_id': transaction.id})
            + f'?transaction_id={transaction.order_id}&origin=origin&ok=1&signed=1'
        ).follow()
        assert Payment.objects.all().count() == 1
        payment = Payment.objects.all().first()
        assert payment.amount == 10
        assert payment.payment_type.slug == 'creditcard'
        assert payment.get_invoice_payments()[0].invoice.paid_amount == 10
        assert payment.get_invoice_payments()[0].invoice.remaining_amount == 0
        assert {x[0][0].url for x in mock_send.call_args_list} == {'http://payment1.com/'}


def test_basket_payment_with_assigned_credits(app, simple_user):
    app = login(app, username='user', password='user')

    regie = Regie.objects.create(label='Foo')
    PaymentType.create_defaults(regie)
    PaymentBackend.objects.create(label='Test', service='dummy', regie=regie)
    invoice = DraftInvoice.objects.create(
        regie=regie,
        date_publication=datetime.date(2023, 4, 21),
        date_payment_deadline=datetime.date(2023, 4, 22),
        date_due=datetime.date(2023, 4, 23),
    )
    DraftInvoiceLine.objects.create(
        slug='event-a-foo-bar',
        label='Event A',
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice,
        quantity=10,
        unit_amount=1,
        user_external_id='user:1',
    )

    basket = Basket.objects.create(
        regie=regie,
        draft_invoice=invoice,
        payer_nameid='ab' * 16,
        payer_external_id='payer:1',
        expiry_at=now() + datetime.timedelta(hours=1),
    )
    line = BasketLine.objects.create(
        basket=basket,
        closed=False,
        user_external_id='user:1',
        payment_callback_url='http://payment1.com',
    )
    BasketLineItem.objects.create(
        line=line,
        label='Event A',
        subject='Réservation',
        details='Lun 06/11, Mar 07/11',
        quantity=10,
        unit_amount=1,
    )
    line.closed = True
    line.save()

    credit1 = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:1',
    )
    CreditLine.objects.create(
        credit=credit1,
        event_date=datetime.date(2022, 9, 1),
        quantity=1,
        unit_amount=1,
    )
    credit2 = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:1',
    )
    CreditLine.objects.create(
        credit=credit2,
        event_date=datetime.date(2022, 9, 1),
        quantity=3,
        unit_amount=1,
    )
    other_regie = Regie.objects.create(label='Foo')
    other_credit = Credit.objects.create(
        regie=other_regie,  # other regie
        payer_external_id='payer:1',
    )
    CreditLine.objects.create(
        credit=other_credit,
        event_date=datetime.date(2022, 9, 1),
        quantity=3,
        unit_amount=1,
    )
    other_credit = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:2',  # other payer
    )
    CreditLine.objects.create(
        credit=other_credit,
        event_date=datetime.date(2022, 9, 1),
        quantity=3,
        unit_amount=1,
    )

    resp = app.get('/basket/validate/')
    resp = resp.form.submit()
    assert resp.location.startswith('https://dummy-payment.entrouvert.com/')
    assert Transaction.objects.all().count() == 1
    transaction = Transaction.objects.all().first()
    assert transaction.status == 0
    assert transaction.amount == 6

    with mock.patch('lingo.utils.requests_wrapper.RequestsSession.send') as mock_send:
        app.get(
            reverse('lingo-epayment-explicit-return', kwargs={'transaction_id': transaction.id})
            + f'?transaction_id={transaction.order_id}&origin=origin&ok=1&signed=1'
        ).follow()
    assert Payment.objects.count() == 2
    payment1, payment2 = Payment.objects.all().order_by('pk')
    assert payment1.amount == 6
    assert payment1.payment_type.slug == 'creditcard'
    assert payment1.get_invoice_payments()[0].invoice.paid_amount == 10
    assert payment1.get_invoice_payments()[0].invoice.remaining_amount == 0
    assert payment2.amount == 4
    assert payment2.payment_type.slug == 'credit'
    assert payment2.get_invoice_payments()[0].invoice.paid_amount == 10
    assert payment2.get_invoice_payments()[0].invoice.remaining_amount == 0
    assert {x[0][0].url for x in mock_send.call_args_list} == {'http://payment1.com/'}
    assert payment1.get_invoice_payments()[0].invoice == payment2.get_invoice_payments()[0].invoice
    assert CreditAssignment.objects.count() == 2
    assignment1, assignment2 = CreditAssignment.objects.all().order_by('pk')
    assert assignment1.amount == 1
    assert assignment1.invoice == payment2.get_invoice_payments()[0].invoice
    assert assignment1.payment == payment2
    assert assignment1.credit == credit1
    assert assignment2.amount == 3
    assert assignment2.invoice == payment2.get_invoice_payments()[0].invoice
    assert assignment2.payment == payment2
    assert assignment2.credit == credit2
