import base64
import datetime
import decimal
from urllib.parse import urlparse

import pytest
from django.contrib.auth.models import Group
from django.urls import reverse
from django.utils.formats import date_format
from django.utils.timezone import localtime, now
from pyquery import PyQuery
from webtest import Upload

from lingo.agendas.models import Agenda
from lingo.invoicing.models import (
    DEFAULT_PAYMENT_TYPES,
    AppearanceSettings,
    Campaign,
    Counter,
    Credit,
    CreditAssignment,
    CreditLine,
    DraftJournalLine,
    InjectedLine,
    Invoice,
    InvoiceLine,
    InvoiceLinePayment,
    JournalLine,
    Payer,
    Payment,
    PaymentType,
    Pool,
    Refund,
    Regie,
)
from tests.utils import login

pytestmark = pytest.mark.django_db


def test_manager_invoicing_regie_list_title(app, admin_user):
    app = login(app)
    resp = app.get(reverse('lingo-manager-invoicing-regie-list'))
    h2 = resp.pyquery('div#appbar h2')
    assert h2.text() == 'Regies'


def test_manager_invoicing_regie_list_empty(app, admin_user):
    app = login(app)
    resp = app.get(reverse('lingo-manager-invoicing-regie-list'))
    msg_info = resp.pyquery('div#content div.big-msg-info')
    assert (
        msg_info.text()
        == "This site doesn't have any regie yet. Click on the \"New\" button in the top right of the page to add a first one."
    )


def test_manager_invoicing_regie_list_add_button(app, admin_user):
    app = login(app)
    resp = app.get(reverse('lingo-manager-invoicing-regie-list'))
    add_button = resp.pyquery('a[href="%s"]' % reverse('lingo-manager-invoicing-regie-add'))
    assert add_button.text() == 'New regie'


def test_manager_invoicing_regie_list_show_objetcs(app, admin_user):
    app = login(app)
    regie = Regie.objects.create(label='Foo')
    resp = app.get(reverse('lingo-manager-invoicing-regie-list'))
    href = resp.pyquery(
        'div#content div ul li a[href="%s"]'
        % reverse('lingo-manager-invoicing-regie-detail', kwargs={'pk': regie.pk})
    )
    assert href.text() == 'Foo [identifier: foo]'


def test_manager_invoicing_regie_add(app, admin_user):
    app = login(app)
    assert Regie.objects.count() == 0
    group = Group.objects.create(name='role-foo')
    resp = app.get(reverse('lingo-manager-invoicing-regie-add'))
    h2 = resp.pyquery('div#appbar h2')
    assert h2.text() == 'New regie'
    form = resp.form
    form.set('label', 'Foo')
    form.set('description', 'foo description')
    form.set('cashier_role', group.id)
    response = form.submit().follow()
    assert Regie.objects.count() == 1
    regie = Regie.objects.first()
    assert regie.label == 'Foo'
    assert regie.slug == 'foo'
    assert regie.description == 'foo description'
    assert regie.cashier_role == group
    assert regie.paymenttype_set.count() == len(DEFAULT_PAYMENT_TYPES)
    assert urlparse(response.request.url).path == reverse(
        'lingo-manager-invoicing-regie-detail', kwargs={'pk': regie.pk}
    )


def test_manager_invoicing_regie_detail(app, admin_user):
    app = login(app)
    group = Group.objects.create(name='role-foo')
    regie = Regie.objects.create(label='Foo', description='foo description', cashier_role=group)
    resp = app.get(reverse('lingo-manager-invoicing-regie-detail', kwargs={'pk': regie.pk}))
    h2 = resp.pyquery('div#appbar h2')
    assert h2.text() == 'Regie - Foo'
    campaign = Campaign.objects.create(
        regie=regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
    )
    resp = app.get(reverse('lingo-manager-invoicing-regie-detail', kwargs={'pk': regie.pk}))
    href = resp.pyquery(
        'ul li a[href="%s"]'
        % reverse('lingo-manager-invoicing-campaign-detail', kwargs={'regie_pk': regie.pk, 'pk': campaign.pk})
    )
    assert href.text() == '(01/09/2022 - 01/10/2022) [due date: 31/10/2022]'

    campaign.finalized = True
    campaign.save()
    resp = app.get(reverse('lingo-manager-invoicing-regie-detail', kwargs={'pk': regie.pk}))
    href = resp.pyquery(
        'ul li a[href="%s"]'
        % reverse('lingo-manager-invoicing-campaign-detail', kwargs={'regie_pk': regie.pk, 'pk': campaign.pk})
    )
    assert href.text() == '(01/09/2022 - 01/10/2022) [due date: 31/10/2022] validated'


def test_manager_invoicing_regie_parameters(app, admin_user):
    app = login(app)
    group = Group.objects.create(name='role-foo')
    regie = Regie.objects.create(label='Foo', description='foo description', cashier_role=group)
    resp = app.get(reverse('lingo-manager-invoicing-regie-parameters', kwargs={'pk': regie.pk}))
    h2 = resp.pyquery('div#appbar h2')
    assert h2.text() == 'Regie - Foo'
    descr = resp.pyquery('div#panel-settings p')[0]
    assert descr.text == 'foo description'
    slug = resp.pyquery('div#panel-settings ul li')[0]
    assert slug.text == 'Identifier: foo'
    slug = resp.pyquery('div#panel-settings ul li')[1]
    assert slug.text == 'Payer: '
    cashier_role = resp.pyquery('div#panel-settings ul li')[2]
    assert cashier_role.text == 'Cashier role: role-foo'
    usage = resp.pyquery('div#panel-usage div')[0]
    assert 'This Regie is not used yet.' in usage.text
    edit_button = resp.pyquery(
        'a[href="%s"]' % reverse('lingo-manager-invoicing-regie-edit', kwargs={'pk': regie.pk})
    )
    assert edit_button.text() == 'Edit'
    delete_button = resp.pyquery(
        'a[href="%s"]' % reverse('lingo-manager-invoicing-regie-delete', kwargs={'pk': regie.pk})
    )
    assert delete_button.text() == 'Delete'

    agenda1 = Agenda.objects.create(label='Foo Bar', regie=regie)
    agenda2 = Agenda.objects.create(label='Foo Bar 2', regie=regie)
    agenda3 = Agenda.objects.create(label='Foo Bar 3')
    resp = app.get(reverse('lingo-manager-invoicing-regie-parameters', kwargs={'pk': regie.pk}))
    assert '/manage/pricing/agenda/%s/' % agenda1.pk in resp
    assert '/manage/pricing/agenda/%s/' % agenda2.pk in resp
    assert '/manage/pricing/agenda/%s/' % agenda3.pk not in resp


def test_manager_invoicing_regie_edit(app, admin_user):
    app = login(app)
    group_foo = Group.objects.create(name='role-foo')
    group_bar = Group.objects.create(name='role-bar')
    regie = Regie.objects.create(label='Foo', description='foo description', cashier_role=group_foo)
    resp = app.get(reverse('lingo-manager-invoicing-regie-edit', kwargs={'pk': regie.pk}))
    h2 = resp.pyquery('div#appbar h2')
    assert h2.text() == 'Edit regie - Foo'
    form = resp.form
    form.set('label', 'Foo bar')
    form.set('slug', 'foo-bar')
    form.set('description', 'foo new description')
    form.set('cashier_role', group_bar.id)
    response = form.submit().follow()
    assert Regie.objects.count() == 1
    regie = Regie.objects.first()
    assert regie.label == 'Foo bar'
    assert regie.slug == 'foo-bar'
    assert regie.description == 'foo new description'
    assert regie.cashier_role == group_bar
    assert urlparse(response.request.url).path == reverse(
        'lingo-manager-invoicing-regie-parameters', kwargs={'pk': regie.pk}
    )

    Regie.objects.create(label='Foo', description='foo description', cashier_role=group_foo)
    resp = app.get(reverse('lingo-manager-invoicing-regie-edit', kwargs={'pk': regie.pk}))
    form = resp.form
    form.set('slug', 'foo')
    response = form.submit()
    assert response.context['form'].errors['slug'] == ['Another regie exists with the same identifier.']

    payer = Payer.objects.create(label='Foo Bar')
    resp = app.get(reverse('lingo-manager-invoicing-regie-edit', kwargs={'pk': regie.pk}))
    form = resp.form
    form.set('payer', payer.pk)
    resp = resp.form.submit().follow()
    regie.refresh_from_db()
    assert regie.payer == payer
    assert 'Payer: <a href="/manage/invoicing/payer/%s/">Foo Bar</a>' % payer.pk in resp


def test_manager_invoicing_regie_counters_edit(app, admin_user):
    app = login(app)
    regie = Regie.objects.create(label='Foo', description='foo description')
    resp = app.get(reverse('lingo-manager-invoicing-regie-counters-edit', kwargs={'pk': regie.pk}))
    h2 = resp.pyquery('div#appbar h2')
    assert h2.text() == 'Edit counters'
    form = resp.form
    form.set('counter_name', '{yyyy}')
    form.set('invoice_number_format', 'Ffoobar-{yyyy}-{number:08d}')
    form.set('payment_number_format', 'Rfoobar-{yyyy}-{number:08d}')
    form.set('credit_number_format', 'Afoobar-{yyyy}-{number:08d}')
    form.set('refund_number_format', 'Vfoobar-{yyyy}-{number:08d}')
    resp = form.submit()
    regie.refresh_from_db()
    assert regie.counter_name == '{yyyy}'
    assert regie.invoice_number_format == 'Ffoobar-{yyyy}-{number:08d}'
    assert regie.payment_number_format == 'Rfoobar-{yyyy}-{number:08d}'
    assert regie.credit_number_format == 'Afoobar-{yyyy}-{number:08d}'
    assert regie.refund_number_format == 'Vfoobar-{yyyy}-{number:08d}'
    assert resp.location.endswith('/manage/invoicing/regie/%s/parameters/#open:counters' % regie.pk)


def test_manager_invoicing_appearance_settings(app, admin_user, settings):
    app = login(app)
    regie = Regie.objects.create(label='Foo', description='foo description')
    assert regie.invoice_model == 'middle'
    resp = app.get(reverse('lingo-manager-invoicing-regie-list'))
    resp = resp.click('Appearance Settings')
    resp.form['logo'] = Upload(
        'test.png',
        base64.decodebytes(
            b'iVBORw0KGgoAAAANSUhEUgAAAAEAAAABAQAAAAA3bvkkAAAACklEQVQI12NoAAAAggCB3UNq9AAAAABJRU5ErkJggg=='
        ),
        'image/png',
    )
    resp.form['address'] = '<p>Foo bar<br>Streetname</p>'
    resp.form['extra_info'] = '<p>Opening hours...</p>'
    resp = resp.form.submit('submit').follow()
    appearance_settings = AppearanceSettings.singleton()
    assert appearance_settings.logo.name == 'logo/test.png'
    assert appearance_settings.address == '<p>Foo bar<br>Streetname</p>'
    assert appearance_settings.extra_info == '<p>Opening hours...</p>'
    resp = resp.click('Appearance Settings')
    assert resp.form['address'].value == '<p>Foo bar<br>Streetname</p>'

    # regie settings
    resp = app.get(reverse('lingo-manager-invoicing-regie-publishing-edit', kwargs={'pk': regie.pk}))
    resp.form.set('invoice_custom_text', '<p>custom text</p>')
    assert resp.form['invoice_model'].options == [
        ('basic', False, 'Basic'),
        ('middle', True, 'Middle'),
        ('full', False, 'Full'),
    ]
    resp.form['invoice_model'] = 'basic'
    resp.form['invoice_main_colour'] = '#9141ac'
    resp = resp.form.submit()
    regie.refresh_from_db()
    assert regie.invoice_custom_text == '<p>custom text</p>'
    assert regie.invoice_model == 'basic'
    assert regie.invoice_main_colour == '#9141ac'
    assert resp.location.endswith('/manage/invoicing/regie/%s/parameters/#open:publishing' % regie.pk)

    # check French typography fixes
    settings.LANGUAGE_CODE = 'fr-fr'
    resp = app.get(reverse('lingo-manager-invoicing-regie-publishing-edit', kwargs={'pk': regie.pk}))
    resp.form.set('invoice_custom_text', '<p>custom : text</p>')
    resp = resp.form.submit()
    regie.refresh_from_db()
    assert regie.invoice_custom_text == '<p>custom\u00a0: text</p>'


def test_manager_invoicing_regie_delete(app, admin_user):
    app = login(app)
    group = Group.objects.create(name='role-foo')
    regie = Regie.objects.create(label='Foo', description='foo description', cashier_role=group)
    assert Regie.objects.count() == 1
    resp = app.get(reverse('lingo-manager-invoicing-regie-delete', kwargs={'pk': regie.pk}))
    response = resp.form.submit().follow()
    assert Regie.objects.count() == 0
    assert urlparse(response.request.url).path == reverse('lingo-manager-invoicing-regie-list')

    # can not delete regie containing campaign
    regie.save()
    campaign = Campaign.objects.create(
        regie=regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
    )
    resp = app.get(reverse('lingo-manager-invoicing-regie-parameters', kwargs={'pk': regie.pk}))
    assert reverse('lingo-manager-invoicing-regie-delete', kwargs={'pk': regie.pk}) not in resp
    resp = app.get(reverse('lingo-manager-invoicing-regie-delete', kwargs={'pk': regie.pk}), status=404)

    campaign.delete()

    # can not delete regie containing injected line
    injected_line = InjectedLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        amount=1,
        regie=regie,
    )
    resp = app.get(reverse('lingo-manager-invoicing-regie-parameters', kwargs={'pk': regie.pk}))
    assert reverse('lingo-manager-invoicing-regie-delete', kwargs={'pk': regie.pk}) not in resp
    resp = app.get(reverse('lingo-manager-invoicing-regie-delete', kwargs={'pk': regie.pk}), status=404)

    injected_line.delete()

    # check counters are deleted
    Counter.get_count(regie=regie, name='bar', kind='invoice')
    Counter.get_count(regie=regie, name='bar', kind='payment')
    Counter.get_count(regie=regie, name='bar', kind='credit')
    Counter.get_count(regie=regie, name='bar', kind='refund')
    assert Counter.objects.count() == 4
    resp = app.get(reverse('lingo-manager-invoicing-regie-delete', kwargs={'pk': regie.pk}))
    response = resp.form.submit().follow()
    assert Regie.objects.count() == 0
    assert Counter.objects.count() == 0


def test_add_payment_type(app, admin_user):
    regie = Regie.objects.create(label='Regie')

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/parameters/' % regie.pk)
    resp = resp.click('New payment type')
    resp.form['label'] = 'Foo'
    assert 'slug' not in resp.context['form'].fields
    assert 'disabled' not in resp.context['form'].fields
    resp = resp.form.submit()
    payment_type = PaymentType.objects.latest('pk')
    assert resp.location.endswith('/manage/invoicing/regie/%s/parameters/#open:payment-types' % regie.pk)
    assert payment_type.label == 'Foo'
    assert payment_type.regie == regie
    assert payment_type.slug == 'foo'
    assert payment_type.disabled is False

    resp = app.get('/manage/invoicing/regie/%s/payment-type/add/' % regie.pk)
    resp.form['label'] = 'Foo'
    resp = resp.form.submit()
    assert resp.location.endswith('/manage/invoicing/regie/%s/parameters/#open:payment-types' % regie.pk)
    payment_type = PaymentType.objects.latest('pk')
    assert payment_type.label == 'Foo'
    assert payment_type.regie == regie
    assert payment_type.slug == 'foo-1'
    assert payment_type.disabled is False


def test_edit_payment_type(app, admin_user):
    regie = Regie.objects.create(label='Regie')
    payment_type = PaymentType.objects.create(label='Foo', regie=regie)
    payment_type2 = PaymentType.objects.create(label='Baz', regie=regie)
    regie2 = Regie.objects.create(label='Regie2')
    payment_type3 = PaymentType.objects.create(label='Foo bar', regie=regie2)

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/parameters/' % regie.pk)
    resp = resp.click(href='/manage/invoicing/regie/%s/payment-type/%s/edit/' % (regie.pk, payment_type.pk))
    assert 'This check type is set on some existing payments, modify it with caution.' not in resp
    resp.form['label'] = 'Foo bar'
    resp.form['slug'] = payment_type2.slug
    resp.form['disabled'] = True
    resp = resp.form.submit()
    assert resp.context['form'].errors['slug'] == ['Another payment type exists with the same identifier.']

    resp.form['slug'] = payment_type3.slug
    resp = resp.form.submit()
    assert resp.location.endswith('/manage/invoicing/regie/%s/parameters/#open:payment-types' % regie.pk)
    payment_type.refresh_from_db()
    assert payment_type.label == 'Foo bar'
    assert payment_type.slug == 'foo-bar'
    assert payment_type.disabled is True

    app.get('/manage/invoicing/regie/%s/payment-type/%s/edit/' % (regie2.pk, payment_type.pk), status=404)


def test_delete_payment_type(app, admin_user):
    regie = Regie.objects.create(label='Regie')
    regie2 = Regie.objects.create(label='Regie2')
    payment_type = PaymentType.objects.create(label='Foo', regie=regie)
    payment = Payment.objects.create(
        regie=regie,
        amount=1,
        payment_type=payment_type,
    )

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/parameters/' % regie.pk)
    assert '/manage/invoicing/regie/%s/payment-type/%s/delete/' % (regie.pk, payment_type.pk) not in resp
    app.get('/manage/invoicing/regie/%s/payment-type/%s/delete/' % (regie.pk, payment_type.pk), status=404)

    payment.delete()

    app.get('/manage/invoicing/regie/%s/payment-type/%s/delete/' % (regie2.pk, payment_type.pk), status=404)

    resp = app.get('/manage/invoicing/regie/%s/parameters/' % regie.pk)
    resp = resp.click(href='/manage/invoicing/regie/%s/payment-type/%s/delete/' % (regie.pk, payment_type.pk))
    resp = resp.form.submit()
    assert resp.location.endswith('/manage/invoicing/regie/%s/parameters/#open:payment-types' % regie.pk)
    assert PaymentType.objects.exists() is False


def test_non_invoiced_line_list(app, admin_user):
    regie = Regie.objects.create(label='Regie')
    other_regie = Regie.objects.create(label='Other Regie')

    campaign = Campaign.objects.create(
        regie=regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
    )
    pool = Pool.objects.create(
        campaign=campaign,
        draft=True,
    )
    campaign2 = Campaign.objects.create(
        regie=regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
    )
    pool2 = Pool.objects.create(
        campaign=campaign2,
        draft=False,
    )

    other_campaign = Campaign.objects.create(
        regie=other_regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
    )
    other_pool = Pool.objects.create(
        campaign=other_campaign,
        draft=True,
    )
    other_campaign2 = Campaign.objects.create(
        regie=other_regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
    )
    other_pool2 = Pool.objects.create(
        campaign=other_campaign2,
        draft=False,
    )

    # not invoiced
    InjectedLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        slug='event-2022-09-01',
        label='Event 2022-09-01',
        amount=3,
        user_external_id='user:1',
        payer_external_id='payer:1',
        regie=regie,
    )
    # not invoiced but in another regie
    InjectedLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        slug='other-event-2022-09-01',
        label='Other Event 2022-09-01',
        amount=3,
        user_external_id='user:1',
        payer_external_id='payer:1',
        regie=other_regie,
    )
    # not invoiced, but linked in a DraftJournalLine
    injected_line2 = InjectedLine.objects.create(
        event_date=datetime.date(2022, 9, 2),
        slug='event-2022-09-02',
        label='Event 2022-09-02',
        amount=3,
        user_external_id='user:1',
        payer_external_id='payer:1',
        regie=regie,
    )
    DraftJournalLine.objects.create(
        event_date=datetime.date(2022, 9, 2),
        slug='event-2022-09-02',
        label='Event 2022-09-02',
        amount=3,
        pool=pool,
        from_injected_line=injected_line2,
    )
    # not invoiced, but linked in a DraftJournalLine, but in another regie
    other_injected_line2 = InjectedLine.objects.create(
        event_date=datetime.date(2022, 9, 2),
        slug='other-event-2022-09-02',
        label='Other Event 2022-09-02',
        amount=3,
        user_external_id='user:1',
        payer_external_id='payer:1',
        regie=other_regie,
    )
    DraftJournalLine.objects.create(
        event_date=datetime.date(2022, 9, 2),
        slug='other-event-2022-09-02',
        label='Other Event 2022-09-02',
        amount=3,
        pool=other_pool,
        from_injected_line=other_injected_line2,
    )
    # invoiced, as linked in a non draft pool
    injected_line3 = InjectedLine.objects.create(
        event_date=datetime.date(2022, 9, 3),
        slug='event-2022-09-03',
        label='Event 2022-09-03',
        amount=3,
        user_external_id='user:1',
        payer_external_id='payer:1',
        regie=regie,
    )
    JournalLine.objects.create(
        event_date=datetime.date(2022, 9, 3),
        slug='event-2022-09-03',
        label='Event 2022-09-03',
        amount=3,
        pool=pool2,
        from_injected_line=injected_line3,
    )

    # non fixed error
    JournalLine.objects.create(
        event_date=datetime.date(2022, 9, 4),
        slug='event-2022-09-04',
        label='Event 2022-09-04',
        amount=0,
        pool=pool2,
        status='error',
    )
    # non fixed error, but in another regie
    JournalLine.objects.create(
        event_date=datetime.date(2022, 9, 4),
        slug='other-event-2022-09-04',
        label='Other Event 2022-09-04',
        amount=0,
        pool=other_pool2,
        status='error',
    )
    # fixed or ignored errors
    JournalLine.objects.create(
        event_date=datetime.date(2022, 9, 5),
        slug='event-2022-09-05',
        label='Event 2022-09-05',
        amount=0,
        pool=pool2,
        status='error',
        error_status='fixed',
    )
    JournalLine.objects.create(
        event_date=datetime.date(2022, 9, 6),
        slug='event-2022-09-06',
        label='Event 2022-09-06',
        amount=0,
        pool=pool2,
        status='error',
        error_status='ignored',
    )
    # not errors
    JournalLine.objects.create(
        event_date=datetime.date(2022, 9, 7),
        slug='event-2022-09-07',
        label='Event 2022-09-07',
        amount=0,
        pool=pool2,
        status='success',
    )
    JournalLine.objects.create(
        event_date=datetime.date(2022, 9, 8),
        slug='event-2022-09-08',
        label='Event 2022-09-08',
        amount=0,
        pool=pool2,
        status='warning',
    )

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/non-invoiced-lines/' % regie.pk)
    assert 'event-2022-09-01' in resp
    assert 'other-event-2022-09-01' not in resp
    assert 'event-2022-09-02' in resp
    assert 'other-event-2022-09-02' not in resp
    assert 'event-2022-09-03' not in resp
    assert 'event-2022-09-04' in resp
    assert 'other-event-2022-09-04' not in resp
    assert 'event-2022-09-05' not in resp
    assert 'event-2022-09-06' not in resp
    assert 'event-2022-09-07' not in resp
    assert 'event-2022-09-08' not in resp


def test_non_invoiced_line_link(app, admin_user, settings):
    settings.KNOWN_SERVICES = {}

    regie = Regie.objects.create(label='Regie')
    campaign = Campaign.objects.create(
        regie=regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
    )
    pool = Pool.objects.create(
        campaign=campaign,
        draft=False,
    )

    line = JournalLine.objects.create(
        event_date=datetime.date(2022, 9, 4),
        slug='event-2022-09-04',
        label='Event 2022-09-04',
        amount=0,
        pool=pool,
        status='error',
    )

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/non-invoiced-lines/' % regie.pk)
    assert 'see agenda' not in resp
    assert 'see event' not in resp

    line.event = {
        'agenda': 'foobar',
    }
    line.save()
    resp = app.get('/manage/invoicing/regie/%s/non-invoiced-lines/' % regie.pk)
    assert '<a href="/manage/pricing/agenda/foobar/">see agenda</a>' in resp
    assert 'see event' not in resp

    line.event['slug'] = 'bazbaz'
    line.save()
    resp = app.get('/manage/invoicing/regie/%s/non-invoiced-lines/' % regie.pk)
    assert '<a href="/manage/pricing/agenda/foobar/">see agenda</a>' in resp
    assert 'see event' not in resp

    settings.KNOWN_SERVICES['chrono'] = {'default': {'url': 'https://chrono.dev/'}}
    resp = app.get('/manage/invoicing/regie/%s/non-invoiced-lines/' % regie.pk)
    assert '<a href="/manage/pricing/agenda/foobar/">see agenda</a>' in resp
    assert '<a href="https://chrono.dev/manage/agendas/foobar/events/bazbaz/">see event</a>' in resp


@pytest.mark.parametrize('orphan', [True, False])
def test_regie_invoices(app, admin_user, orphan):
    regie = Regie.objects.create(label='Foo')
    Agenda.objects.create(label='Agenda A', regie=regie)
    Agenda.objects.create(label='Agenda B', regie=regie)
    PaymentType.create_defaults(regie)
    today = now().date()
    tomorrow = today + datetime.timedelta(days=1)
    yesterday = today - datetime.timedelta(days=1)
    if not orphan:
        campaign1 = Campaign.objects.create(
            regie=regie,
            date_start=datetime.date(2022, 9, 1),
            date_end=datetime.date(2022, 10, 1),
            date_publication=datetime.date(2022, 10, 1),
            date_payment_deadline=datetime.date(2022, 10, 31),
            date_due=datetime.date(2022, 10, 31),
            date_debit=datetime.date(2022, 11, 15),
            finalized=True,
        )
        pool1 = Pool.objects.create(
            campaign=campaign1,
            draft=False,
            status='completed',
        )
        campaign2 = Campaign.objects.create(
            regie=regie,
            date_start=datetime.date(2022, 9, 1),
            date_end=datetime.date(2022, 10, 1),
            date_publication=datetime.date(2022, 10, 1),
            date_payment_deadline=datetime.date(2022, 10, 31),
            date_due=datetime.date(2022, 10, 31),
            date_debit=datetime.date(2022, 11, 15),
            finalized=True,
        )
        pool2 = Pool.objects.create(
            campaign=campaign2,
            draft=False,
            status='completed',
        )
    invoice1 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        regie=regie,
        pool=pool1 if not orphan else None,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=True,
        payer_direct_debit=False,
    )
    invoice1.set_number()
    invoice1.save()
    invoice2 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
        regie=regie,
        pool=pool2 if not orphan else None,
        payer_external_id='payer:2',
        payer_first_name='First2',
        payer_last_name='Name2',
        payer_address='42 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=False,
        payer_direct_debit=True,
    )
    invoice2.set_number()
    invoice2.save()
    invoice3 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
        regie=regie,
        pool=pool2 if not orphan else None,
        payer_external_id='payer:3',
        payer_first_name='First3',
        payer_last_name='Name3',
        payer_address='43 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=False,
        payer_direct_debit=True,
    )
    invoice3.set_number()
    invoice3.save()
    invoice4 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        regie=regie,
        cancelled_at=now(),
    )
    invoice4.set_number()
    invoice4.save()

    invoice_line11 = InvoiceLine.objects.create(
        slug='event-a-foo-bar',
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice1,
        quantity=1.2,
        unit_amount=1,
        pool=pool1 if not orphan else None,
        label='Event A',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
        details={
            'agenda': 'agenda-a',
            'primary_event': 'event-a',
            'status': 'presence',
            'check_type': 'foo',
            'check_type_group': 'foobar',
            'check_type_label': 'Foo!',
            'dates': ['2022-09-01', '2022-09-02', '2022-09-03'],
        },
        event_slug='agenda-a@event-a',
        agenda_slug='agenda-a',
        activity_label='Agenda A',
        description='Thu01, Fri02, Sat03',
    )
    invoice_line12 = InvoiceLine.objects.create(
        slug='event-b-foo-bar',
        event_date=datetime.date(2022, 9, 2),
        invoice=invoice1,
        quantity=1,
        unit_amount=2,
        pool=pool1 if not orphan else None,
        label='Event B',
        user_external_id='user:2',
        user_first_name='User2',
        user_last_name='Name2',
        details={
            'agenda': 'agenda-b',
            'primary_event': 'event-b',
            'status': 'presence',
            'check_type': 'foo',
            'check_type_group': 'foobar',
            'check_type_label': 'Foo!',
            'dates': ['2022-09-01', '2022-09-02', '2022-09-03'],
        },
        event_slug='agenda-b@event-b',
        agenda_slug='agenda-b',
        activity_label='Agenda B',
        description='Thu01, Fri02, Sat03',
    )
    invoice_line13 = InvoiceLine.objects.create(
        slug='event-a-foo-bar',
        event_date=datetime.date(2022, 9, 3),
        invoice=invoice1,
        quantity=1,
        unit_amount=3,
        pool=pool1 if not orphan else None,
        label='Event A',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
        details={
            'agenda': 'agenda-a',
            'primary_event': 'event-a',
            'status': 'presence',
            'dates': ['2022-09-04', '2022-09-05'],
        },
        event_slug='agenda-a@event-a',
        agenda_slug='agenda-a',
        activity_label='Agenda A',
        description='Sun04, Mon05',
    )
    payment1 = Payment.objects.create(
        regie=regie,
        amount=1,
        payment_type=PaymentType.objects.get(regie=regie, slug='cash'),
    )
    payment1.set_number()
    payment1.save()
    invoice_payment1 = InvoiceLinePayment.objects.create(
        payment=payment1,
        line=invoice_line11,
        amount=1,
    )
    invoice1.refresh_from_db()
    assert invoice1.remaining_amount == decimal.Decimal('5.2')
    assert invoice1.paid_amount == 1

    invoice_line21 = InvoiceLine.objects.create(
        slug='agenda-a@event-aa',  # non recurring event
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice2,
        quantity=1,
        unit_amount=1,
        pool=pool2 if not orphan else None,
        label='Event AA',
        event_slug='agenda-a@event-aa',
        agenda_slug='agenda-a',
        activity_label='Agenda A',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    payment2 = Payment.objects.create(
        regie=regie,
        amount=0.5,
        payment_type=PaymentType.objects.get(regie=regie, slug='cash'),
    )
    payment2.set_number()
    payment2.save()
    invoice_payment2 = InvoiceLinePayment.objects.create(
        payment=payment2,
        line=invoice_line21,
        amount=0.5,
    )
    payment3 = Payment.objects.create(
        regie=regie,
        amount=0.5,
        payment_type=PaymentType.objects.get(regie=regie, slug='cesu'),
    )
    payment3.set_number()
    payment3.save()
    invoice_payment3 = InvoiceLinePayment.objects.create(
        payment=payment3,
        line=invoice_line21,
        amount=0.5,
    )
    invoice2.refresh_from_db()
    assert invoice2.remaining_amount == 0
    assert invoice2.paid_amount == 1

    invoice_line31 = InvoiceLine.objects.create(
        slug='injected',  # from injected line
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice3,
        quantity=1,
        unit_amount=1,
        pool=pool2 if not orphan else None,
        label='Event A',
        event_slug='injected',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    invoice3.refresh_from_db()
    assert invoice3.remaining_amount == 1
    assert invoice3.paid_amount == 0

    non_finalized_campaign = Campaign.objects.create(
        regie=regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
    )
    non_finalized_pool = Pool.objects.create(
        campaign=non_finalized_campaign,
        draft=False,
        status='completed',
    )
    non_finalized_invoice = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        regie=regie,
        pool=non_finalized_pool,
        payer_external_id='payer:1',
    )
    non_finalized_invoice.set_number()
    non_finalized_invoice.save()
    InvoiceLine.objects.create(
        slug='event-a-foo-bar',
        event_date=datetime.date(2022, 9, 1),
        invoice=non_finalized_invoice,
        quantity=1.2,
        unit_amount=1,
        pool=non_finalized_pool,
        label='Event A',
        event_slug='event-a-foo-bar',
        user_external_id='user:1',
        description='@overtaking@',
    )

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/invoices/' % regie.pk)
    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % invoice1.pk
    ).text() == 'Partially paid Invoice F%02d-%s-0000001 dated %s addressed to First1 Name1, amount 6.20€ - download' % (
        regie.pk,
        invoice1.created_at.strftime('%y-%m'),
        invoice1.created_at.strftime('%d/%m/%Y'),
    )
    if orphan:
        assert len(resp.pyquery('tr[data-invoicing-element-id="%s"] a' % invoice1.pk)) == 2
    else:
        assert len(resp.pyquery('tr[data-invoicing-element-id="%s"] a' % invoice1.pk)) == 3
    lines_url = resp.pyquery('tr[data-invoicing-element-id="%s"]' % invoice1.pk).attr(
        'data-invoicing-element-lines-url'
    )
    assert lines_url == '/manage/invoicing/ajax/regie/%s/invoice/%s/lines/' % (
        regie.pk,
        invoice1.pk,
    )
    lines_resp = app.get(lines_url)
    assert len(lines_resp.pyquery('tr')) == 28
    assert [PyQuery(tr).text() for tr in lines_resp.pyquery('tr')] == [
        'Demat: yes',
        'Direct debit: no',
        'Publication date: 01/10/2022',
        'Payment deadline: 31/10/2022',
        'Due date: 31/10/2022',
        'User1 Name1',
        'Description\nAmount\nQuantity\nSubtotal',
        'Agenda A',
        'Event A\nFoo! Thu01, Fri02, Sat03\n1.00€\n1.2\n1.20€',
        'Payments',
        'Payment\nDate\nType\nAmount',
        'R%02d-%s-0000001\n%s\nCash\n1.00€'
        % (
            regie.pk,
            payment1.created_at.strftime('%y-%m'),
            date_format(localtime(invoice_payment1.created_at), 'DATETIME_FORMAT'),
        ),
        'Event A\nPresence Sun04, Mon05\n3.00€\n1\n3.00€',
        'Payments',
        'Payment\nDate\nType\nAmount',
        'No payments for this line',
        'User2 Name2',
        'Description\nAmount\nQuantity\nSubtotal',
        'Agenda B',
        'Event B\nFoo! Thu01, Fri02, Sat03\n2.00€\n1\n2.00€',
        'Payments',
        'Payment\nDate\nType\nAmount',
        'No payments for this line',
        'Payments',
        'Payment\nDate\nType\nAmount',
        'R%02d-%s-0000001\n%s\nCash\n1.00€'
        % (
            regie.pk,
            payment1.created_at.strftime('%y-%m'),
            date_format(localtime(invoice_payment1.created_at), 'DATETIME_FORMAT'),
        ),
        'Paid amount: 1.00€',
        'Remaining amount: 5.20€',
    ]
    if not orphan:
        assert [PyQuery(a).attr('href') for a in lines_resp.pyquery('tr a')] == [
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?user_external_id=user:1'
            % (regie.pk, campaign1.pk, pool1.pk),
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?invoice_line=%s'
            % (regie.pk, campaign1.pk, pool1.pk, invoice_line11.pk),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000001'
            % (
                regie.pk,
                regie.pk,
                payment1.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?invoice_line=%s'
            % (regie.pk, campaign1.pk, pool1.pk, invoice_line13.pk),
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?user_external_id=user:2'
            % (regie.pk, campaign1.pk, pool1.pk),
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?invoice_line=%s'
            % (regie.pk, campaign1.pk, pool1.pk, invoice_line12.pk),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000001'
            % (
                regie.pk,
                regie.pk,
                payment1.created_at.strftime('%y-%m'),
            ),
        ]
    else:
        assert [PyQuery(a).attr('href') for a in lines_resp.pyquery('tr a')] == [
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000001'
            % (
                regie.pk,
                regie.pk,
                payment1.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000001'
            % (
                regie.pk,
                regie.pk,
                payment1.created_at.strftime('%y-%m'),
            ),
        ]
    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % invoice2.pk
    ).text() == 'Paid Invoice F%02d-%s-0000002 dated %s addressed to First2 Name2, amount 1.00€ - download' % (
        regie.pk,
        invoice2.created_at.strftime('%y-%m'),
        invoice2.created_at.strftime('%d/%m/%Y'),
    )
    lines_url = resp.pyquery('tr[data-invoicing-element-id="%s"]' % invoice2.pk).attr(
        'data-invoicing-element-lines-url'
    )
    assert lines_url == '/manage/invoicing/ajax/regie/%s/invoice/%s/lines/' % (
        regie.pk,
        invoice2.pk,
    )
    lines_resp = app.get(lines_url)
    assert len(lines_resp.pyquery('tr')) == 20
    assert [PyQuery(tr).text() for tr in lines_resp.pyquery('tr')] == [
        'Demat: no',
        'Direct debit: no',
        'Publication date: 01/10/2022',
        'Payment deadline: 31/10/2022',
        'Due date: 31/10/2022',
        'Debit date: 15/11/2022',
        'User1 Name1',
        'Description\nAmount\nQuantity\nSubtotal',
        'Agenda A',
        'Event AA\n1.00€\n1\n1.00€',
        'Payments',
        'Payment\nDate\nType\nAmount',
        'R%02d-%s-0000002\n%s\nCash\n0.50€'
        % (
            regie.pk,
            payment2.created_at.strftime('%y-%m'),
            date_format(localtime(invoice_payment2.created_at), 'DATETIME_FORMAT'),
        ),
        'R%02d-%s-0000003\n%s\nCESU\n0.50€'
        % (
            regie.pk,
            payment3.created_at.strftime('%y-%m'),
            date_format(localtime(invoice_payment3.created_at), 'DATETIME_FORMAT'),
        ),
        'Payments',
        'Payment\nDate\nType\nAmount',
        'R%02d-%s-0000002\n%s\nCash\n0.50€'
        % (
            regie.pk,
            payment2.created_at.strftime('%y-%m'),
            date_format(localtime(invoice_payment2.created_at), 'DATETIME_FORMAT'),
        ),
        'R%02d-%s-0000003\n%s\nCESU\n0.50€'
        % (
            regie.pk,
            payment3.created_at.strftime('%y-%m'),
            date_format(localtime(invoice_payment3.created_at), 'DATETIME_FORMAT'),
        ),
        'Paid amount: 1.00€',
        'Payments certificate: download',
    ]
    if not orphan:
        assert [PyQuery(a).attr('href') for a in lines_resp.pyquery('tr a')] == [
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?user_external_id=user:1'
            % (regie.pk, campaign2.pk, pool2.pk),
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?invoice_line=%s'
            % (regie.pk, campaign2.pk, pool2.pk, invoice_line21.pk),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000002'
            % (
                regie.pk,
                regie.pk,
                payment2.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000003'
            % (
                regie.pk,
                regie.pk,
                payment3.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000002'
            % (
                regie.pk,
                regie.pk,
                payment2.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000003'
            % (
                regie.pk,
                regie.pk,
                payment3.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/invoice/%s/payments/pdf/' % (regie.pk, invoice2.pk),
        ]
    else:
        assert [PyQuery(a).attr('href') for a in lines_resp.pyquery('tr a')] == [
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000002'
            % (
                regie.pk,
                regie.pk,
                payment2.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000003'
            % (
                regie.pk,
                regie.pk,
                payment3.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000002'
            % (
                regie.pk,
                regie.pk,
                payment2.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000003'
            % (
                regie.pk,
                regie.pk,
                payment3.created_at.strftime('%y-%m'),
            ),
            '/manage/invoicing/regie/%s/invoice/%s/payments/pdf/' % (regie.pk, invoice2.pk),
        ]

    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % invoice3.pk
    ).text() == 'Invoice F%02d-%s-0000003 dated %s addressed to First3 Name3, amount 1.00€ - download' % (
        regie.pk,
        invoice3.created_at.strftime('%y-%m'),
        invoice3.created_at.strftime('%d/%m/%Y'),
    )
    if orphan:
        assert len(resp.pyquery('tr[data-invoicing-element-id="%s"] a' % invoice3.pk)) == 2
    else:
        assert len(resp.pyquery('tr[data-invoicing-element-id="%s"] a' % invoice3.pk)) == 3
    lines_url = resp.pyquery('tr[data-invoicing-element-id="%s"]' % invoice3.pk).attr(
        'data-invoicing-element-lines-url'
    )
    assert lines_url == '/manage/invoicing/ajax/regie/%s/invoice/%s/lines/' % (
        regie.pk,
        invoice3.pk,
    )
    lines_resp = app.get(lines_url)
    assert len(lines_resp.pyquery('tr')) == 16
    assert [PyQuery(tr).text() for tr in lines_resp.pyquery('tr')] == [
        'Demat: no',
        'Direct debit: no',
        'Publication date: 01/10/2022',
        'Payment deadline: 31/10/2022',
        'Due date: 31/10/2022',
        'Debit date: 15/11/2022',
        'User1 Name1',
        'Description\nAmount\nQuantity\nSubtotal',
        'Event A\n1.00€\n1\n1.00€',
        'Payments',
        'Payment\nDate\nType\nAmount',
        'No payments for this line',
        'Payments',
        'Payment\nDate\nType\nAmount',
        'No payments for this invoice',
        'Remaining amount: 1.00€',
    ]
    if not orphan:
        assert [PyQuery(a).attr('href') for a in lines_resp.pyquery('tr a')] == [
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?user_external_id=user:1'
            % (regie.pk, campaign2.pk, pool2.pk),
            '/manage/invoicing/regie/%s/campaign/%s/pool/%s/journal/?invoice_line=%s'
            % (regie.pk, campaign2.pk, pool2.pk, invoice_line31.pk),
        ]

    resp = app.get('/manage/invoicing/regie/%s/invoices/?csv' % regie.pk)
    assert resp.headers['Content-Type'] == 'text/csv'
    assert resp.headers['Content-Disposition'] == 'attachment; filename="invoices.csv"'
    assert len([a for a in resp.text.split('\r\n') if a]) == 1 + 3
    assert resp.text == (
        'Number,Payer ID,Payer first name,Payer last name,Publication date,Payment deadline,Due date,Demat,Direct debit,Total due,Paid amount,Status\r\n'
        'F%02d-%s-0000003,payer:3,First3,Name3,2022-10-01,2022-10-31,2022-10-31,no,yes,1.00,0.00,Not paid\r\n'
        'F%02d-%s-0000002,payer:2,First2,Name2,2022-10-01,2022-10-31,2022-10-31,no,yes,1.00,1.00,Paid\r\n'
        'F%02d-%s-0000001,payer:1,First1,Name1,2022-10-01,2022-10-31,2022-10-31,yes,no,6.20,1.00,Partially paid\r\n'
    ) % (
        regie.pk,
        invoice3.created_at.strftime('%y-%m'),
        regie.pk,
        invoice2.created_at.strftime('%y-%m'),
        regie.pk,
        invoice1.created_at.strftime('%y-%m'),
    )

    resp = app.get('/manage/invoicing/regie/%s/invoices/?csv&full' % regie.pk)
    assert resp.headers['Content-Type'] == 'text/csv'
    assert resp.headers['Content-Disposition'] == 'attachment; filename="invoices-full.csv"'
    assert len([a for a in resp.text.split('\r\n') if a]) == 1 + 5
    assert resp.text == (
        'Number,Payer ID,Payer first name,Payer last name,Publication date,Payment deadline,Due date,Demat,Direct debit,'
        'Description,Unit amount,Quantity,Total due,Payment type,Paid amount,Status\r\n'
        'F%02d-%s-0000003,payer:3,First3,Name3,2022-10-01,2022-10-31,2022-10-31,no,yes,Event A,1.00,1.00,1.00,,0.00,Not paid\r\n'
        'F%02d-%s-0000002,payer:2,First2,Name2,2022-10-01,2022-10-31,2022-10-31,no,yes,Event AA,1.00,1.00,1.00,"CESU, Cash",1.00,Paid\r\n'
        'F%02d-%s-0000001,payer:1,First1,Name1,2022-10-01,2022-10-31,2022-10-31,yes,no,Event A,1.00,1.20,1.20,Cash,1.00,Partially paid\r\n'
        'F%02d-%s-0000001,payer:1,First1,Name1,2022-10-01,2022-10-31,2022-10-31,yes,no,Event B,2.00,1.00,2.00,,0.00,Not paid\r\n'
        'F%02d-%s-0000001,payer:1,First1,Name1,2022-10-01,2022-10-31,2022-10-31,yes,no,Event A,3.00,1.00,3.00,,0.00,Not paid\r\n'
    ) % (
        regie.pk,
        invoice3.created_at.strftime('%y-%m'),
        regie.pk,
        invoice2.created_at.strftime('%y-%m'),
        regie.pk,
        invoice1.created_at.strftime('%y-%m'),
        regie.pk,
        invoice1.created_at.strftime('%y-%m'),
        regie.pk,
        invoice1.created_at.strftime('%y-%m'),
    )

    # test filters
    params = [
        ({'number': invoice1.formatted_number}, 1),
        ({'number': invoice1.created_at.strftime('%y-%m')}, 3),
        ({'created_at_after': today.strftime('%Y-%m-%d')}, 3),
        ({'created_at_after': tomorrow.strftime('%Y-%m-%d')}, 0),
        ({'created_at_before': yesterday.strftime('%Y-%m-%d')}, 0),
        ({'created_at_before': today.strftime('%Y-%m-%d')}, 3),
        ({'date_payment_deadline_after': '2022-10-31'}, 3),
        ({'date_payment_deadline_after': '2022-11-01'}, 0),
        ({'date_payment_deadline_before': '2022-10-30'}, 0),
        ({'date_payment_deadline_before': '2022-10-31'}, 3),
        ({'date_due_after': '2022-10-31'}, 3),
        ({'date_due_after': '2022-11-01'}, 0),
        ({'date_due_before': '2022-10-30'}, 0),
        ({'date_due_before': '2022-10-31'}, 3),
        ({'payment_number': payment1.formatted_number}, 1),
        ({'payment_number': payment1.created_at.strftime('%y-%m')}, 2),
        ({'payer_external_id': 'payer:1'}, 1),
        ({'payer_external_id': 'payer:2'}, 1),
        ({'payer_first_name': 'first'}, 3),
        ({'payer_first_name': 'first1'}, 1),
        ({'payer_last_name': 'name'}, 3),
        ({'payer_last_name': 'name1'}, 1),
        ({'payer_demat': True}, 1),
        ({'payer_direct_debit': True}, 2),
        ({'user_external_id': 'user:1'}, 3),
        ({'user_external_id': 'user:2'}, 1),
        ({'user_first_name': 'user'}, 3),
        ({'user_first_name': 'user2'}, 1),
        ({'user_last_name': 'name'}, 3),
        ({'user_last_name': 'name1'}, 3),
        (
            {
                'total_amount_min': '1',
                'total_amount_min_lookup': 'gt',
            },
            1,
        ),
        (
            {
                'total_amount_min': '1',
                'total_amount_min_lookup': 'gte',
            },
            3,
        ),
        (
            {
                'total_amount_max': '6.2',
                'total_amount_max_lookup': 'lt',
            },
            2,
        ),
        (
            {
                'total_amount_max': '6.2',
                'total_amount_max_lookup': 'lte',
            },
            3,
        ),
        ({'paid': 'yes'}, 1),
        ({'paid': 'partially'}, 1),
        ({'paid': 'no'}, 1),
        ({'agenda': 'agenda-a'}, 2),
        ({'agenda': 'agenda-b'}, 1),
        ({'event': 'agenda-a@event-a'}, 1),
        ({'event': 'agenda-a@event-aa'}, 1),
        ({'event': 'agenda-b@event-b'}, 1),
    ]
    for param, result in params:
        resp = app.get(
            '/manage/invoicing/regie/%s/invoices/' % regie.pk,
            params=param,
        )
        assert len(resp.pyquery('tr.invoice')) == result
        param['csv'] = True
        resp = app.get(
            '/manage/invoicing/regie/%s/invoices/' % regie.pk,
            params=param,
        )
        assert len([a for a in resp.text.split('\r\n') if a]) == 1 + result

    app.get(
        '/manage/invoicing/ajax/regie/%s/invoice/%s/lines/' % (regie.pk, non_finalized_invoice.pk), status=404
    )


def test_regie_invoice_pdf(app, admin_user):
    regie = Regie.objects.create(label='Foo', invoice_main_colour='#9141ac')
    campaign = Campaign.objects.create(
        regie=regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
        finalized=True,
    )
    assert campaign.invoice_model == 'middle'
    pool = Pool.objects.create(
        campaign=campaign,
        draft=False,
        status='completed',
    )
    invoice = Invoice.objects.create(
        label='Invoice from 01/09/2022 to 30/09/2022',
        date_publication=campaign.date_publication,
        date_payment_deadline=campaign.date_payment_deadline,
        date_due=campaign.date_due,
        regie=regie,
        pool=pool,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=True,
        payer_direct_debit=False,
    )
    invoice.set_number()
    invoice.save()

    InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice,
        quantity=1.2,
        unit_amount=1,
        pool=pool,
        label='Label 11',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
        details={
            'agenda': 'agenda-1',
            'primary_event': 'event-1',
            'status': 'presence',
            'check_type': 'foo',
            'check_type_group': 'foobar',
            'check_type_label': 'Foo!',
            'dates': ['2022-09-01', '2022-09-02', '2022-09-03'],
        },
        event_slug='agenda-1@event-1',
        agenda_slug='agenda-1',
        activity_label='Agenda 1',
        description='Thu01, Fri02, Sat03',
    )
    InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 2),
        invoice=invoice,
        quantity=1,
        unit_amount=2,
        pool=pool,
        label='Label 12',
        user_external_id='user:2',
        user_first_name='User2',
        user_last_name='Name2',
    )
    InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 3),
        invoice=invoice,
        quantity=1,
        unit_amount=3,
        pool=pool,
        label='Label 13',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
        description='@overtaking@',
    )
    InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice,
        quantity=15,
        unit_amount=1,
        pool=pool,
        label='Agenda foobar',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
        details={
            'agenda': 'agenda-2',
            'primary_event': 'event-1',
            'status': 'presence',
            'partial_bookings': True,
        },
        event_slug='agenda-2@event-1',
        agenda_slug='agenda-2',
        activity_label='Agenda 2',
        description='a description',
    )
    InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice,
        quantity=0,
        unit_amount=1,
        pool=pool,
        label='Absence',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
        details={
            'agenda': 'agenda-2',
            'primary_event': 'event-1',
            'status': 'absence',
            'partial_bookings': True,
        },
        event_slug='agenda-2@event-1',
        agenda_slug='agenda-2',
        activity_label='Agenda 2',
        description='a description',
    )

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (regie.pk, invoice.pk))
    assert 'color: #9141ac;' in resp
    assert resp.pyquery('#document-label').text() == 'Invoice from 01/09/2022 to 30/09/2022'
    assert resp.pyquery('#regie-label').text() == 'Foo'
    assert resp.pyquery('address#to').text() == 'First1 Name1\n41 rue des kangourous\n99999 Kangourou Ville'
    assert resp.pyquery('dl#informations').text() == 'Invoice number:\nF%02d-%s-0000001\nDate:\n%s' % (
        regie.pk,
        invoice.created_at.strftime('%y-%m'),
        date_format(localtime(invoice.created_at), 'DATE_FORMAT'),
    )
    assert [PyQuery(tr).text() for tr in resp.pyquery('table#lines tr')] == [
        'User1 Name1',
        'Services\nDetails\nUA\nQTY\nTA',
        'Label 11\nAgenda 1\nFoo!\nThu01, Fri02, Sat03\n1.00€\n1.2\n1.20€',
        'Agenda foobar\na description\n1.00€\n15\n15.00€',
        'Absence\na description\n1.00€\n0\n0.00€',
        'Label 13\n\n3.00€\n1\n3.00€',
        'User2 Name2',
        'Services\nDetails\nUA\nQTY\nTA',
        'Label 12\n\n2.00€\n1\n2.00€',
        'Total amount to be paid before: Oct. 31, 2022\n21.20€',
    ]
    assert len(resp.pyquery('table#lines-details')) == 0

    campaign.invoice_model = 'basic'
    campaign.save()
    resp = app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (regie.pk, invoice.pk))
    assert [PyQuery(tr).text() for tr in resp.pyquery('table#lines tr')] == [
        'User1 Name1',
        'Services\nUA\nQTY\nTA',
        'Label 11\nAgenda 1\nFoo!\n1.00€\n1.2\n1.20€',
        'Agenda foobar\n1.00€\n15\n15.00€',
        'Absence\n1.00€\n0\n0.00€',
        'Label 13\n\n3.00€\n1\n3.00€',
        'User2 Name2',
        'Services\nUA\nQTY\nTA',
        'Label 12\n\n2.00€\n1\n2.00€',
        'Total amount to be paid before: Oct. 31, 2022\n21.20€',
    ]
    assert len(resp.pyquery('table#lines-details')) == 0

    campaign.invoice_model = 'full'
    campaign.save()
    resp = app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (regie.pk, invoice.pk))
    assert [PyQuery(tr).text() for tr in resp.pyquery('table#lines tr')] == [
        'User1 Name1',
        'Services\nUA\nQTY\nTA',
        'Label 11\nAgenda 1\nFoo!\n1.00€\n1.2\n1.20€',
        'Agenda foobar\n1.00€\n15\n15.00€',
        'Absence\n1.00€\n0\n0.00€',
        'Label 13\n\n3.00€\n1\n3.00€',
        'User2 Name2',
        'Services\nUA\nQTY\nTA',
        'Label 12\n\n2.00€\n1\n2.00€',
        'Total amount to be paid before: Oct. 31, 2022\n21.20€',
    ]
    assert len(resp.pyquery('table#lines-details')) == 1
    assert [PyQuery(tr).text() for tr in resp.pyquery('table#lines-details tr')] == [
        'User1 Name1',
        'Services\nDetails',
        'Label 11\nAgenda 1\nFoo!\nThu01, Fri02, Sat03',
        'Agenda foobar\na description',
        'Absence\na description',
    ]

    app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (0, invoice.pk), status=404)
    app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (regie.pk, 0), status=404)
    other_regie = Regie.objects.create(label='Foo')
    app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (other_regie.pk, invoice.pk), status=404)

    appearance_settings = AppearanceSettings.singleton()
    appearance_settings.address = '<p>Foo bar<br>Streetname</p>'
    appearance_settings.extra_info = '<p>Opening hours...</p>'
    appearance_settings.save()
    resp = app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (regie.pk, invoice.pk))
    assert appearance_settings.address in resp.text
    assert appearance_settings.extra_info in resp.text

    campaign.finalized = False
    campaign.save()
    app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (regie.pk, invoice.pk), status=404)
    invoice.pool = None
    invoice.save()
    app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (regie.pk, invoice.pk), status=200)
    invoice.cancelled_at = now()
    invoice.save()
    app.get('/manage/invoicing/regie/%s/invoice/%s/pdf/?html' % (regie.pk, invoice.pk), status=404)


def test_regie_invoice_payments_pdf(app, admin_user):
    regie = Regie.objects.create(
        label='Foo',
        cashier_name='Le régisseur principal',
        city_name='Kangourou Ville',
        invoice_main_colour='#9141ac',
    )
    campaign = Campaign.objects.create(
        regie=regie,
        date_start=datetime.date(2022, 9, 1),
        date_end=datetime.date(2022, 10, 1),
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
        finalized=True,
    )
    pool = Pool.objects.create(
        campaign=campaign,
        draft=False,
        status='completed',
    )
    PaymentType.create_defaults(regie)
    invoice = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=True,
        payer_direct_debit=False,
        label='Invoice Label',
    )
    invoice.set_number()
    invoice.save()

    line = InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice,
        quantity=1,
        unit_amount=40,
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )

    payment1 = Payment.objects.create(
        regie=regie,
        amount=5,
        payment_type=PaymentType.objects.get(regie=regie, slug='check'),
    )
    payment1.set_number()
    payment1.save()
    InvoiceLinePayment.objects.create(
        payment=payment1,
        line=line,
        amount=5,
    )
    payment2 = Payment.objects.create(
        regie=regie,
        amount=35,
        payment_type=PaymentType.objects.get(regie=regie, slug='online'),
    )
    payment2.set_number()
    payment2.save()
    InvoiceLinePayment.objects.create(
        payment=payment2,
        line=line,
        amount=35,
    )
    invoice.refresh_from_db()
    assert invoice.remaining_amount == 0

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/invoice/%s/payments/pdf/?html' % (regie.pk, invoice.pk))
    assert 'color: #9141ac;' in resp
    assert resp.pyquery('#document-label').text() == 'Payments certificate'
    assert resp.pyquery('.address-to-container').text() == (
        'Invoiced account:\nFirst1 Name1 (1)\nInvoicing address:\n41 rue des kangourous\n99999 Kangourou Ville\n'
        'Invoice number:\nF%02d-%s-0000001\nInvoice object:\nInvoice Label'
    ) % (
        regie.pk,
        invoice.created_at.strftime('%y-%m'),
    )
    assert (
        resp.pyquery('p#informations').text()
        == 'Hereby certifies that I have received the amount of 40.00€ for account 1 First1 Name1 as follows:'
    )
    assert [PyQuery(tr).text() for tr in resp.pyquery.find('thead tr')] == [
        'Number\nDate\nPayment type\nAmount',
    ]
    assert [PyQuery(tr).text() for tr in resp.pyquery.find('tbody tr')] == [
        'R%02d-%s-0000001\n%s\nCheck\n5.00€'
        % (
            regie.pk,
            payment1.created_at.strftime('%y-%m'),
            payment1.created_at.strftime('%d/%m/%Y'),
        ),
        'R%02d-%s-0000002\n%s\nOnline\n35.00€'
        % (
            regie.pk,
            payment2.created_at.strftime('%y-%m'),
            payment2.created_at.strftime('%d/%m/%Y'),
        ),
    ]
    assert resp.pyquery(
        '#regie-signature'
    ).text() == 'Le régisseur principal\nKangourou Ville, on %s' % invoice.created_at.strftime('%d/%m/%Y')

    app.get('/manage/invoicing/regie/%s/invoice/%s/payments/pdf/?html' % (0, invoice.pk), status=404)
    app.get('/manage/invoicing/regie/%s/invoice/%s/payments/pdf/?html' % (regie.pk, 0), status=404)
    other_regie = Regie.objects.create(label='Foo')
    app.get(
        '/manage/invoicing/regie/%s/invoice/%s/payments/pdf/?html' % (other_regie.pk, invoice.pk), status=404
    )
    line = InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice,
        quantity=1,
        unit_amount=40,
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    invoice.refresh_from_db()
    assert invoice.remaining_amount > 0
    app.get('/manage/invoicing/regie/%s/invoice/%s/payments/pdf/?html' % (regie.pk, invoice.pk), status=404)

    line.delete()
    invoice.refresh_from_db()
    assert invoice.remaining_amount == 0
    invoice.pool = pool
    invoice.save()
    app.get('/manage/invoicing/regie/%s/invoice/%s/payments/pdf/?html' % (regie.pk, invoice.pk), status=200)
    campaign.finalized = False
    campaign.save()
    app.get('/manage/invoicing/regie/%s/invoice/%s/payments/pdf/?html' % (regie.pk, invoice.pk), status=404)
    campaign.finalized = True
    campaign.save()
    invoice.cancelled_at = now()
    invoice.save()
    app.get('/manage/invoicing/regie/%s/invoice/%s/payments/pdf/?html' % (regie.pk, invoice.pk), status=404)


def test_regie_payments(app, admin_user):
    regie = Regie.objects.create(label='Foo')
    Agenda.objects.create(label='Agenda A', regie=regie)
    Agenda.objects.create(label='Agenda B', regie=regie)
    PaymentType.create_defaults(regie)
    invoice1 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=True,
        payer_direct_debit=False,
    )
    invoice1.set_number()
    invoice1.save()
    invoice2 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=False,
        payer_direct_debit=True,
    )
    invoice2.set_number()
    invoice2.save()
    invoice3 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
        regie=regie,
        payer_external_id='payer:3',
        payer_first_name='First3',
        payer_last_name='Name3',
        payer_address='43 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=False,
        payer_direct_debit=True,
    )
    invoice3.set_number()
    invoice3.save()

    invoice_line1 = InvoiceLine.objects.create(
        slug='event-a-foo-bar',
        label='Event A',
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice1,
        quantity=1,
        unit_amount=40,
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
        details={
            'agenda': 'agenda-a',
            'primary_event': 'event-a',
        },
        event_slug='agenda-a@event-a',
        agenda_slug='agenda-a',
        activity_label='Agenda A',
    )
    invoice_line2 = InvoiceLine.objects.create(
        slug='agenda-b@event-b',  # non recurring event
        label='Event B',
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice2,
        quantity=1,
        unit_amount=50,
        event_slug='agenda-b@event-b',
        agenda_slug='agenda-b',
        activity_label='Agenda B',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    invoice_line3 = InvoiceLine.objects.create(
        slug='injected',
        label='Event A',
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice3,
        quantity=1,
        unit_amount=60,
        event_slug='injected',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )

    payment1 = Payment.objects.create(
        regie=regie,
        amount=35,
        payment_type=PaymentType.objects.get(regie=regie, slug='cash'),
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payment_info={
            'check_number': '123456',
        },
    )
    payment1.set_number()
    payment1.save()
    InvoiceLinePayment.objects.create(
        payment=payment1,
        line=invoice_line1,
        amount=35,
    )

    payment2 = Payment.objects.create(
        regie=regie,
        amount=55,
        payment_type=PaymentType.objects.get(regie=regie, slug='check'),
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payment_info={
            'check_number': '123456',
            'check_issuer': 'Foo',
            'check_bank': 'Bar',
            'payment_reference': 'Ref',
        },
    )
    payment2.set_number()
    payment2.save()
    InvoiceLinePayment.objects.create(
        payment=payment2,
        line=invoice_line1,
        amount=5,
    )
    InvoiceLinePayment.objects.create(
        payment=payment2,
        line=invoice_line2,
        amount=50,
    )

    payment3 = Payment.objects.create(
        regie=regie,
        amount=2,
        payment_type=PaymentType.objects.get(regie=regie, slug='creditcard'),
        payer_external_id='payer:3',
        payer_first_name='First3',
        payer_last_name='Name3',
        payer_address='43 rue des kangourous\n99999 Kangourou Ville',
    )
    payment3.set_number()
    payment3.save()
    InvoiceLinePayment.objects.create(
        payment=payment3,
        line=invoice_line3,
        amount=2,
    )

    invoice1.refresh_from_db()
    assert invoice1.remaining_amount == 0
    assert invoice1.paid_amount == 40
    invoice2.refresh_from_db()
    assert invoice2.remaining_amount == 0
    assert invoice2.paid_amount == 50
    invoice3.refresh_from_db()
    assert invoice3.remaining_amount == 58
    assert invoice3.paid_amount == 2

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/payments/' % regie.pk)
    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % payment1.pk
    ).text() == 'Payment R%02d-%s-0000001 dated %s from First1 Name1, amount 35.00€ (Cash) - download' % (
        regie.pk,
        payment1.created_at.strftime('%y-%m'),
        payment1.created_at.strftime('%d/%m/%Y'),
    )
    assert [
        PyQuery(tr).text() for tr in resp.pyquery('tr[data-related-invoicing-element-id="%s"]' % payment1.pk)
    ] == [
        'Invoice\nAmount charged\nAmount assigned',
        'F%02d-%s-0000001\n40.00€\n35.00€'
        % (
            regie.pk,
            invoice1.created_at.strftime('%y-%m'),
        ),
        'Number: 123456',
    ]
    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % payment2.pk
    ).text() == 'Payment R%02d-%s-0000002 dated %s from First1 Name1, amount 55.00€ (Check) - download' % (
        regie.pk,
        payment2.created_at.strftime('%y-%m'),
        payment2.created_at.strftime('%d/%m/%Y'),
    )
    assert [
        PyQuery(tr).text() for tr in resp.pyquery('tr[data-related-invoicing-element-id="%s"]' % payment2.pk)
    ] == [
        'Invoice\nAmount charged\nAmount assigned',
        'F%02d-%s-0000001\n40.00€\n5.00€'
        % (
            regie.pk,
            invoice1.created_at.strftime('%y-%m'),
        ),
        'F%02d-%s-0000002\n50.00€\n50.00€'
        % (
            regie.pk,
            invoice2.created_at.strftime('%y-%m'),
        ),
        'Issuer: Foo',
        'Bank/Organism: Bar',
        'Number: 123456',
        'Reference: Ref',
    ]
    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % payment3.pk
    ).text() == 'Payment R%02d-%s-0000003 dated %s from First3 Name3, amount 2.00€ (Credit card) - download' % (
        regie.pk,
        payment3.created_at.strftime('%y-%m'),
        payment3.created_at.strftime('%d/%m/%Y'),
    )
    assert [
        PyQuery(tr).text() for tr in resp.pyquery('tr[data-related-invoicing-element-id="%s"]' % payment3.pk)
    ] == [
        'Invoice\nAmount charged\nAmount assigned',
        'F%02d-%s-0000003\n60.00€\n2.00€'
        % (
            regie.pk,
            invoice3.created_at.strftime('%y-%m'),
        ),
    ]

    resp = app.get('/manage/invoicing/regie/%s/payments/?csv' % regie.pk)
    assert resp.headers['Content-Type'] == 'text/csv'
    assert resp.headers['Content-Disposition'] == 'attachment; filename="payments.csv"'
    assert len([a for a in resp.text.split('\r\n') if a]) == 1 + 4
    assert resp.text == (
        'Number,Invoice number,Date,Payer ID,Payer first name,Payer last name,Payment type,Amount assigned,'
        'Total amount,Issuer,Bank/Organism,Number,Reference\r\n'
        'R%02d-%s-0000003,F%02d-%s-0000003,%s,payer:3,First3,Name3,Credit card,2.00,2.00,,,,\r\n'
        'R%02d-%s-0000002,F%02d-%s-0000001,%s,payer:1,First1,Name1,Check,5.00,55.00,Foo,Bar,123456,Ref\r\n'
        'R%02d-%s-0000002,F%02d-%s-0000002,%s,payer:1,First1,Name1,Check,50.00,55.00,Foo,Bar,123456,Ref\r\n'
        'R%02d-%s-0000001,F%02d-%s-0000001,%s,payer:1,First1,Name1,Cash,35.00,35.00,,,123456,\r\n'
    ) % (
        regie.pk,
        payment3.created_at.strftime('%y-%m'),
        regie.pk,
        invoice3.created_at.strftime('%y-%m'),
        payment3.created_at.strftime('%Y-%m-%d'),
        regie.pk,
        payment2.created_at.strftime('%y-%m'),
        regie.pk,
        invoice1.created_at.strftime('%y-%m'),
        payment2.created_at.strftime('%Y-%m-%d'),
        regie.pk,
        payment2.created_at.strftime('%y-%m'),
        regie.pk,
        invoice2.created_at.strftime('%y-%m'),
        payment2.created_at.strftime('%Y-%m-%d'),
        regie.pk,
        payment1.created_at.strftime('%y-%m'),
        regie.pk,
        invoice1.created_at.strftime('%y-%m'),
        payment1.created_at.strftime('%Y-%m-%d'),
    )

    # test filters
    today = datetime.date.today()
    params = [
        ({'number': payment1.formatted_number}, 1, 1),
        ({'number': payment1.created_at.strftime('%y-%m')}, 3, 4),
        ({'created_at_after': today.strftime('%Y-%m-%d')}, 3, 4),
        ({'created_at_after': (today + datetime.timedelta(days=1)).strftime('%Y-%m-%d')}, 0, 0),
        ({'created_at_before': (today - datetime.timedelta(days=1)).strftime('%Y-%m-%d')}, 0, 0),
        ({'created_at_before': today.strftime('%Y-%m-%d')}, 3, 4),
        ({'invoice_number': invoice1.formatted_number}, 2, 3),
        ({'invoice_number': invoice1.created_at.strftime('%y-%m')}, 3, 4),
        ({'payer_external_id': 'payer:1'}, 2, 3),
        ({'payer_external_id': 'payer:3'}, 1, 1),
        ({'payer_first_name': 'first'}, 3, 4),
        ({'payer_first_name': 'first1'}, 2, 3),
        ({'payer_last_name': 'name'}, 3, 4),
        ({'payer_last_name': 'name1'}, 2, 3),
        ({'payment_type': PaymentType.objects.get(slug='cash').pk}, 1, 1),
        ({'payment_type': PaymentType.objects.get(slug='check').pk}, 1, 2),
        (
            {
                'amount_min': '2',
                'amount_min_lookup': 'gt',
            },
            2,
            3,
        ),
        (
            {
                'amount_min': '2',
                'amount_min_lookup': 'gte',
            },
            3,
            4,
        ),
        (
            {
                'amount_max': '55',
                'amount_max_lookup': 'lt',
            },
            2,
            2,
        ),
        (
            {
                'amount_max': '55',
                'amount_max_lookup': 'lte',
            },
            3,
            4,
        ),
        ({'agenda': 'agenda-a'}, 2, 3),
        ({'agenda': 'agenda-b'}, 1, 2),
        ({'event': 'agenda-a@event-a'}, 2, 3),
        ({'event': 'agenda-b@event-b'}, 1, 2),
    ]
    for param, result, csv_result in params:
        resp = app.get(
            '/manage/invoicing/regie/%s/payments/' % regie.pk,
            params=param,
        )
        assert len(resp.pyquery('tr.payment')) == result
        param['csv'] = True
        resp = app.get(
            '/manage/invoicing/regie/%s/payments/' % regie.pk,
            params=param,
        )
        assert len([a for a in resp.text.split('\r\n') if a]) == 1 + csv_result


def test_regie_payment_pdf(app, admin_user):
    regie = Regie.objects.create(
        label='Foo',
        cashier_name='Le régisseur principal',
        city_name='Kangourou Ville',
        invoice_main_colour='#9141ac',
    )
    PaymentType.create_defaults(regie)
    invoice1 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=True,
        payer_direct_debit=False,
    )
    invoice1.set_number()
    invoice1.save()
    invoice2 = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        date_debit=datetime.date(2022, 11, 15),
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        payer_demat=False,
        payer_direct_debit=True,
    )
    invoice2.set_number()
    invoice2.save()

    invoice_line1 = InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice1,
        quantity=1,
        unit_amount=40,
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    invoice_line2 = InvoiceLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        invoice=invoice2,
        quantity=1,
        unit_amount=50,
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )

    payment = Payment.objects.create(
        regie=regie,
        amount=55,
        payment_type=PaymentType.objects.get(regie=regie, slug='check'),
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
    )
    payment.set_number()
    payment.save()
    InvoiceLinePayment.objects.create(
        payment=payment,
        line=invoice_line1,
        amount=5,
    )
    InvoiceLinePayment.objects.create(
        payment=payment,
        line=invoice_line2,
        amount=50,
    )

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/payment/%s/pdf/?html' % (regie.pk, payment.pk))
    assert 'color: #9141ac;' in resp
    assert resp.pyquery('#document-label').text() == 'Payment receipt'
    assert resp.pyquery('.address-to-container').text() == (
        'Invoiced account:\nFirst1 Name1 (1)\nInvoicing address:\n41 rue des kangourous\n99999 Kangourou Ville'
    )
    assert resp.pyquery('p#informations').text() == (
        'Hereby certifies that I have received a payment of type CHECK in the amount of 55.00€ (number R%02d-%s-0000001), for account 1 First1 Name1:'
    ) % (
        regie.pk,
        payment.created_at.strftime('%y-%m'),
    )
    assert [PyQuery(tr).text() for tr in resp.pyquery.find('thead tr')] == [
        'Invoice number\nInvoice object\nAmount charged\nAmount assigned'
    ]
    assert [PyQuery(tr).text() for tr in resp.pyquery.find('tbody tr')] == [
        'F%02d-%s-0000001\n40.00€\n5.00€'
        % (
            regie.pk,
            invoice1.created_at.strftime('%y-%m'),
        ),
        'F%02d-%s-0000002\n50.00€\n50.00€'
        % (
            regie.pk,
            invoice2.created_at.strftime('%y-%m'),
        ),
    ]
    assert resp.pyquery(
        '#regie-signature'
    ).text() == 'Le régisseur principal\nKangourou Ville, on %s' % payment.created_at.strftime('%d/%m/%Y')

    resp = app.get('/manage/invoicing/regie/%s//payment/%s/pdf/?html' % (0, payment.pk), status=404)
    resp = app.get('/manage/invoicing/regie/%s//payment/%s/pdf/?html' % (regie.pk, 0), status=404)
    other_regie = Regie.objects.create(label='Foo')
    resp = app.get(
        '/manage/invoicing/regie/%s/payment/%s/pdf/?html' % (other_regie.pk, payment.pk), status=404
    )


def test_regie_credits(app, admin_user):
    regie = Regie.objects.create(label='Foo')
    Agenda.objects.create(label='Agenda A', regie=regie)
    Agenda.objects.create(label='Agenda B', regie=regie)
    PaymentType.create_defaults(regie)
    invoice = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        regie=regie,
    )
    credit1 = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
    )
    credit1.set_number()
    credit1.save()
    credit2 = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:2',
        payer_first_name='First2',
        payer_last_name='Name2',
        payer_address='42 rue des kangourous\n99999 Kangourou Ville',
    )
    credit2.set_number()
    credit2.save()
    credit3 = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:3',
        payer_first_name='First3',
        payer_last_name='Name3',
        payer_address='43 rue des kangourous\n99999 Kangourou Ville',
    )
    credit3.set_number()
    credit3.save()

    CreditLine.objects.create(
        slug='event-a-foo-bar',
        event_date=datetime.date(2022, 9, 1),
        credit=credit1,
        quantity=1.2,
        unit_amount=1,
        label='Event A',
        description='A description',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    CreditLine.objects.create(
        slug='event-b-foo-bar',
        event_date=datetime.date(2022, 9, 2),
        credit=credit1,
        quantity=1,
        unit_amount=2,
        label='Event B',
        user_external_id='user:2',
        user_first_name='User2',
        user_last_name='Name2',
    )
    CreditLine.objects.create(
        slug='event-a-foo-bar',
        event_date=datetime.date(2022, 9, 3),
        credit=credit1,
        quantity=1,
        unit_amount=3,
        label='Event A',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    payment1 = Payment.objects.create(
        regie=regie,
        amount=1,
        payment_type=PaymentType.objects.get(regie=regie, slug='credit'),
    )
    payment1.set_number()
    payment1.save()
    credit_assignment1 = CreditAssignment.objects.create(
        invoice=invoice,
        payment=payment1,
        credit=credit1,
        amount=1,
    )
    refund = Refund.objects.create(
        regie=regie,
        amount=5.2,
    )
    refund.set_number()
    refund.save()
    credit_assignment2 = CreditAssignment.objects.create(
        refund=refund,
        credit=credit1,
        amount=5.2,
    )
    credit1.refresh_from_db()
    assert credit1.remaining_amount == 0
    assert credit1.assigned_amount == decimal.Decimal('6.2')

    CreditLine.objects.create(
        slug='agenda-a@event-aa',
        event_date=datetime.date(2022, 9, 1),
        credit=credit2,
        quantity=1,
        unit_amount=1,
        label='Event AA',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    CreditAssignment.objects.create(
        invoice=invoice,
        credit=credit2,
        amount=0.5,
    )
    credit2.refresh_from_db()
    assert credit2.remaining_amount == 0.5
    assert credit2.assigned_amount == 0.5

    CreditLine.objects.create(
        slug='injected',
        event_date=datetime.date(2022, 9, 1),
        credit=credit3,
        quantity=1,
        unit_amount=1,
        label='Event A',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    credit3.refresh_from_db()
    assert credit3.remaining_amount == 1
    assert credit3.assigned_amount == 0

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/credits/' % regie.pk)
    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % credit1.pk
    ).text() == 'Assigned Credit A%02d-%s-0000001 dated %s for First1 Name1, amount 6.20€ - download' % (
        regie.pk,
        credit1.created_at.strftime('%y-%m'),
        credit1.created_at.strftime('%d/%m/%Y'),
    )
    assert len(resp.pyquery('tr[data-invoicing-element-id="%s"] a' % credit1.pk)) == 2
    lines_url = resp.pyquery('tr[data-invoicing-element-id="%s"]' % credit1.pk).attr(
        'data-invoicing-element-lines-url'
    )
    assert lines_url == '/manage/invoicing/ajax/regie/%s/credit/%s/lines/' % (
        regie.pk,
        credit1.pk,
    )
    lines_resp = app.get(lines_url)
    assert len(lines_resp.pyquery('tr')) == 12
    assert [PyQuery(tr).text() for tr in lines_resp.pyquery('tr')] == [
        'User1 Name1',
        'Description\nAmount\nQuantity\nSubtotal',
        'Event A\nA description\n1.00€\n1.2\n1.20€',
        'Event A\n3.00€\n1\n3.00€',
        'User2 Name2',
        'Description\nAmount\nQuantity\nSubtotal',
        'Event B\n2.00€\n1\n2.00€',
        'Assignments',
        'Payment\nDate\nAmount',
        'R%02d-%s-0000001\n%s\n1.00€'
        % (
            regie.pk,
            payment1.created_at.strftime('%y-%m'),
            date_format(localtime(credit_assignment1.created_at), 'DATETIME_FORMAT'),
        ),
        'V%02d-%s-0000001 (Refund)\n%s\n5.20€'
        % (
            regie.pk,
            refund.created_at.strftime('%y-%m'),
            date_format(localtime(credit_assignment2.created_at), 'DATETIME_FORMAT'),
        ),
        'Assigned amount: 6.20€',
    ]
    assert [PyQuery(a).attr('href') for a in lines_resp.pyquery('tr a')] == [
        '/manage/invoicing/regie/%s/payments/?number=R%s-%s-0000001'
        % (
            regie.pk,
            regie.pk,
            payment1.created_at.strftime('%y-%m'),
        ),
        '/manage/invoicing/regie/%s/refunds/?number=V%s-%s-0000001'
        % (
            regie.pk,
            regie.pk,
            refund.created_at.strftime('%y-%m'),
        ),
    ]
    assert len(resp.pyquery('tr[data-invoicing-element-id="%s"] a' % credit2.pk)) == 2
    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % credit2.pk
    ).text() == 'Partially assigned Credit A%02d-%s-0000002 dated %s for First2 Name2, amount 1.00€ - download' % (
        regie.pk,
        credit2.created_at.strftime('%y-%m'),
        credit2.created_at.strftime('%d/%m/%Y'),
    )
    lines_url = resp.pyquery('tr[data-invoicing-element-id="%s"]' % credit2.pk).attr(
        'data-invoicing-element-lines-url'
    )
    assert lines_url == '/manage/invoicing/ajax/regie/%s/credit/%s/lines/' % (
        regie.pk,
        credit2.pk,
    )
    lines_resp = app.get(lines_url)
    assert len(lines_resp.pyquery('tr')) == 8
    assert [PyQuery(tr).text() for tr in lines_resp.pyquery('tr')] == [
        'User1 Name1',
        'Description\nAmount\nQuantity\nSubtotal',
        'Event AA\n1.00€\n1\n1.00€',
        'Assignments',
        'Payment\nDate\nAmount',
        'Pending...\n0.50€',
        'Assigned amount: 0.50€',
        'Remaining amount to assign: 0.50€',
    ]
    assert [PyQuery(a).attr('href') for a in lines_resp.pyquery('tr a')] == []

    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % credit3.pk
    ).text() == 'Credit A%02d-%s-0000003 dated %s for First3 Name3, amount 1.00€ - download' % (
        regie.pk,
        credit3.created_at.strftime('%y-%m'),
        credit3.created_at.strftime('%d/%m/%Y'),
    )
    assert len(resp.pyquery('tr[data-invoicing-element-id="%s"] a' % credit3.pk)) == 2
    lines_url = resp.pyquery('tr[data-invoicing-element-id="%s"]' % credit3.pk).attr(
        'data-invoicing-element-lines-url'
    )
    assert lines_url == '/manage/invoicing/ajax/regie/%s/credit/%s/lines/' % (
        regie.pk,
        credit3.pk,
    )
    lines_resp = app.get(lines_url)
    assert len(lines_resp.pyquery('tr')) == 7
    assert [PyQuery(tr).text() for tr in lines_resp.pyquery('tr')] == [
        'User1 Name1',
        'Description\nAmount\nQuantity\nSubtotal',
        'Event A\n1.00€\n1\n1.00€',
        'Assignments',
        'Payment\nDate\nAmount',
        'No assignments for this credit',
        'Remaining amount to assign: 1.00€',
    ]

    # test filters
    today = now().date()
    tomorrow = today + datetime.timedelta(days=1)
    yesterday = today - datetime.timedelta(days=1)
    params = [
        ({'number': credit1.formatted_number}, 1),
        ({'number': credit1.created_at.strftime('%y-%m')}, 3),
        ({'created_at_after': today.strftime('%Y-%m-%d')}, 3),
        ({'created_at_after': tomorrow.strftime('%Y-%m-%d')}, 0),
        ({'created_at_before': yesterday.strftime('%Y-%m-%d')}, 0),
        ({'created_at_before': today.strftime('%Y-%m-%d')}, 3),
        ({'payment_number': payment1.formatted_number}, 1),
        ({'payment_number': payment1.created_at.strftime('%y-%m')}, 1),
        ({'payer_external_id': 'payer:1'}, 1),
        ({'payer_external_id': 'payer:2'}, 1),
        ({'payer_first_name': 'first'}, 3),
        ({'payer_first_name': 'first1'}, 1),
        ({'payer_last_name': 'name'}, 3),
        ({'payer_last_name': 'name1'}, 1),
        ({'user_external_id': 'user:1'}, 3),
        ({'user_external_id': 'user:2'}, 1),
        ({'user_first_name': 'user'}, 3),
        ({'user_first_name': 'user2'}, 1),
        ({'user_last_name': 'name'}, 3),
        ({'user_last_name': 'name1'}, 3),
        (
            {
                'total_amount_min': '1',
                'total_amount_min_lookup': 'gt',
            },
            1,
        ),
        (
            {
                'total_amount_min': '1',
                'total_amount_min_lookup': 'gte',
            },
            3,
        ),
        (
            {
                'total_amount_max': '6.2',
                'total_amount_max_lookup': 'lt',
            },
            2,
        ),
        (
            {
                'total_amount_max': '6.2',
                'total_amount_max_lookup': 'lte',
            },
            3,
        ),
        ({'assigned': 'yes'}, 1),
        ({'assigned': 'partially'}, 1),
        ({'assigned': 'no'}, 1),
    ]
    for param, result in params:
        resp = app.get(
            '/manage/invoicing/regie/%s/credits/' % regie.pk,
            params=param,
        )
        assert len(resp.pyquery('tr.credit')) == result


def test_regie_credit_pdf(app, admin_user):
    regie = Regie.objects.create(label='Foo', invoice_main_colour='#9141ac')
    credit = Credit.objects.create(
        label='Credit from 01/09/2022',
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
    )
    credit.set_number()
    credit.save()

    CreditLine.objects.create(
        event_date=datetime.date(2022, 9, 1),
        credit=credit,
        quantity=1.2,
        unit_amount=1,
        label='Label 11',
        description='A description',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    CreditLine.objects.create(
        event_date=datetime.date(2022, 9, 2),
        credit=credit,
        quantity=1,
        unit_amount=2,
        label='Label 12',
        user_external_id='user:2',
        user_first_name='User2',
        user_last_name='Name2',
    )
    CreditLine.objects.create(
        event_date=datetime.date(2022, 9, 3),
        credit=credit,
        quantity=1,
        unit_amount=3,
        label='Label 13',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/credit/%s/pdf/?html' % (regie.pk, credit.pk))
    assert 'color: #9141ac;' in resp
    assert resp.pyquery('#document-label').text() == 'Credit from 01/09/2022'
    assert resp.pyquery('#regie-label').text() == 'Foo'
    assert resp.pyquery('address#to').text() == 'First1 Name1\n41 rue des kangourous\n99999 Kangourou Ville'
    assert resp.pyquery('dl#informations').text() == 'Credit number:\nA%02d-%s-0000001\nDate:\n%s' % (
        regie.pk,
        credit.created_at.strftime('%y-%m'),
        date_format(localtime(credit.created_at), 'DATE_FORMAT'),
    )
    assert [PyQuery(tr).text() for tr in resp.pyquery('table#lines tr')] == [
        'User1 Name1',
        'Services\nDetails\nUA\nQTY\nTA',
        'Label 11\nA description\n1.00€\n1.2\n1.20€',
        'Label 13\n3.00€\n1\n3.00€',
        'User2 Name2',
        'Services\nDetails\nUA\nQTY\nTA',
        'Label 12\n2.00€\n1\n2.00€',
        'Total amount:\n6.20€',
    ]

    app.get('/manage/invoicing/regie/%s/credit/%s/pdf/?html' % (0, credit.pk), status=404)
    app.get('/manage/invoicing/regie/%s/credit/%s/pdf/?html' % (regie.pk, 0), status=404)
    other_regie = Regie.objects.create(label='Foo')
    app.get('/manage/invoicing/regie/%s/credit/%s/pdf/?html' % (other_regie.pk, credit.pk), status=404)

    appearance_settings = AppearanceSettings.singleton()
    appearance_settings.address = '<p>Foo bar<br>Streetname</p>'
    appearance_settings.extra_info = '<p>Opening hours...</p>'
    appearance_settings.save()
    resp = app.get('/manage/invoicing/regie/%s/credit/%s/pdf/?html' % (regie.pk, credit.pk))
    assert appearance_settings.address in resp.text
    assert appearance_settings.extra_info in resp.text


def test_regie_refunds(app, admin_user):
    regie = Regie.objects.create(label='Foo')
    Agenda.objects.create(label='Agenda A', regie=regie)
    Agenda.objects.create(label='Agenda B', regie=regie)
    PaymentType.create_defaults(regie)
    invoice = Invoice.objects.create(
        date_publication=datetime.date(2022, 10, 1),
        date_payment_deadline=datetime.date(2022, 10, 31),
        date_due=datetime.date(2022, 10, 31),
        regie=regie,
    )
    credit1 = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
    )
    credit1.set_number()
    credit1.save()
    credit2 = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:2',
        payer_first_name='First2',
        payer_last_name='Name2',
        payer_address='42 rue des kangourous\n99999 Kangourou Ville',
    )
    credit2.set_number()
    credit2.save()
    credit3 = Credit.objects.create(
        regie=regie,
        payer_external_id='payer:3',
        payer_first_name='First3',
        payer_last_name='Name3',
        payer_address='43 rue des kangourous\n99999 Kangourou Ville',
    )
    credit3.set_number()
    credit3.save()

    CreditLine.objects.create(
        slug='event-a-foo-bar',
        event_date=datetime.date(2022, 9, 1),
        credit=credit1,
        quantity=1.2,
        unit_amount=1,
        label='Event A',
        description='A description',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    CreditLine.objects.create(
        slug='event-b-foo-bar',
        event_date=datetime.date(2022, 9, 2),
        credit=credit1,
        quantity=1,
        unit_amount=2,
        label='Event B',
        user_external_id='user:2',
        user_first_name='User2',
        user_last_name='Name2',
    )
    CreditLine.objects.create(
        slug='event-a-foo-bar',
        event_date=datetime.date(2022, 9, 3),
        credit=credit1,
        quantity=1,
        unit_amount=3,
        label='Event A',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    refund1 = Refund.objects.create(
        regie=regie,
        payer_external_id='payer:1',
        payer_first_name='First1',
        payer_last_name='Name1',
        payer_address='41 rue des kangourous\n99999 Kangourou Ville',
        amount=6.2,
    )
    refund1.set_number()
    refund1.save()
    CreditAssignment.objects.create(
        refund=refund1,
        credit=credit1,
        amount=6.2,
    )
    credit1.refresh_from_db()
    assert credit1.remaining_amount == 0
    assert credit1.assigned_amount == decimal.Decimal('6.2')

    CreditLine.objects.create(
        slug='agenda-a@event-aa',
        event_date=datetime.date(2022, 9, 1),
        credit=credit2,
        quantity=1,
        unit_amount=1,
        label='Event AA',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    payment2 = Payment.objects.create(
        regie=regie,
        amount=0.5,
        payment_type=PaymentType.objects.get(regie=regie, slug='credit'),
    )
    payment2.set_number()
    payment2.save()
    CreditAssignment.objects.create(
        invoice=invoice,
        payment=payment2,
        credit=credit2,
        amount=0.5,
    )
    refund2 = Refund.objects.create(
        regie=regie,
        payer_external_id='payer:2',
        payer_first_name='First2',
        payer_last_name='Name2',
        payer_address='42 rue des kangourous\n99999 Kangourou Ville',
        amount=0.5,
    )
    refund2.set_number()
    refund2.save()
    CreditAssignment.objects.create(
        refund=refund2,
        credit=credit2,
        amount=0.5,
    )
    credit2.refresh_from_db()
    assert credit2.remaining_amount == 0
    assert credit2.assigned_amount == 1

    CreditLine.objects.create(
        slug='injected',
        event_date=datetime.date(2022, 9, 1),
        credit=credit3,
        quantity=1,
        unit_amount=1,
        label='Event A',
        user_external_id='user:1',
        user_first_name='User1',
        user_last_name='Name1',
    )
    refund3 = Refund.objects.create(
        regie=regie,
        payer_external_id='payer:3',
        payer_first_name='First3',
        payer_last_name='Name3',
        payer_address='43 rue des kangourous\n99999 Kangourou Ville',
        amount=1,
    )
    refund3.set_number()
    refund3.save()
    CreditAssignment.objects.create(
        refund=refund3,
        credit=credit3,
        amount=1,
    )
    credit3.refresh_from_db()
    assert credit3.remaining_amount == 0
    assert credit3.assigned_amount == 1

    app = login(app)
    resp = app.get('/manage/invoicing/regie/%s/refunds/' % regie.pk)
    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % refund1.pk
    ).text() == 'Refund V%02d-%s-0000001 dated %s for First1 Name1, amount 6.20€' % (
        regie.pk,
        refund1.created_at.strftime('%y-%m'),
        refund1.created_at.strftime('%d/%m/%Y'),
    )
    assert [
        PyQuery(tr).text() for tr in resp.pyquery('tr[data-related-invoicing-element-id="%s"]' % refund1.pk)
    ] == [
        'Credit\nDate\nCredit amount\nRefund amount',
        'A%02d-%s-0000001\n%s\n6.20€\n6.20€'
        % (
            regie.pk,
            credit1.created_at.strftime('%y-%m'),
            credit1.created_at.strftime('%d/%m/%Y'),
        ),
    ]

    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % refund2.pk
    ).text() == 'Refund V%02d-%s-0000002 dated %s for First2 Name2, amount 0.50€' % (
        regie.pk,
        refund2.created_at.strftime('%y-%m'),
        refund2.created_at.strftime('%d/%m/%Y'),
    )
    assert [
        PyQuery(tr).text() for tr in resp.pyquery('tr[data-related-invoicing-element-id="%s"]' % refund2.pk)
    ] == [
        'Credit\nDate\nCredit amount\nRefund amount',
        'A%02d-%s-0000002\n%s\n1.00€\n0.50€'
        % (
            regie.pk,
            credit2.created_at.strftime('%y-%m'),
            credit2.created_at.strftime('%d/%m/%Y'),
        ),
    ]

    assert resp.pyquery(
        'tr[data-invoicing-element-id="%s"]' % refund3.pk
    ).text() == 'Refund V%02d-%s-0000003 dated %s for First3 Name3, amount 1.00€' % (
        regie.pk,
        refund3.created_at.strftime('%y-%m'),
        refund3.created_at.strftime('%d/%m/%Y'),
    )
    assert [
        PyQuery(tr).text() for tr in resp.pyquery('tr[data-related-invoicing-element-id="%s"]' % refund3.pk)
    ] == [
        'Credit\nDate\nCredit amount\nRefund amount',
        'A%02d-%s-0000003\n%s\n1.00€\n1.00€'
        % (
            regie.pk,
            credit3.created_at.strftime('%y-%m'),
            credit3.created_at.strftime('%d/%m/%Y'),
        ),
    ]

    # test filters
    today = now().date()
    tomorrow = today + datetime.timedelta(days=1)
    yesterday = today - datetime.timedelta(days=1)
    params = [
        ({'number': refund1.formatted_number}, 1),
        ({'number': refund1.created_at.strftime('%y-%m')}, 3),
        ({'created_at_after': today.strftime('%Y-%m-%d')}, 3),
        ({'created_at_after': tomorrow.strftime('%Y-%m-%d')}, 0),
        ({'created_at_before': yesterday.strftime('%Y-%m-%d')}, 0),
        ({'created_at_before': today.strftime('%Y-%m-%d')}, 3),
        ({'credit_number': credit1.formatted_number}, 1),
        ({'credit_number': credit1.created_at.strftime('%y-%m')}, 3),
        ({'payer_external_id': 'payer:1'}, 1),
        ({'payer_external_id': 'payer:2'}, 1),
        ({'payer_first_name': 'first'}, 3),
        ({'payer_first_name': 'first1'}, 1),
        ({'payer_last_name': 'name'}, 3),
        ({'payer_last_name': 'name1'}, 1),
        (
            {
                'amount_min': '1',
                'amount_min_lookup': 'gt',
            },
            1,
        ),
        (
            {
                'amount_min': '1',
                'amount_min_lookup': 'gte',
            },
            2,
        ),
        (
            {
                'amount_max': '6.2',
                'amount_max_lookup': 'lt',
            },
            2,
        ),
        (
            {
                'amount_max': '6.2',
                'amount_max_lookup': 'lte',
            },
            3,
        ),
    ]
    for param, result in params:
        resp = app.get(
            '/manage/invoicing/regie/%s/refunds/' % regie.pk,
            params=param,
        )
        assert len(resp.pyquery('tr.refund')) == result
